"""This module define class for raster based size function
"""

import functools
import gc
import logging
from multiprocessing import cpu_count, Pool
import operator
import tempfile
from time import time
from typing import Union, List, Callable, Any, Optional, Iterable, Tuple
from contextlib import ExitStack
import warnings
try:
    # pylint: disable=C0412
    from typing import Literal
except ImportError:
    # pylint: disable=C0412
    from typing_extensions import Literal

from jigsawpy import jigsaw_msh_t, jigsaw_jig_t
from jigsawpy import libsaw
import numpy as np
import numpy.typing as npt
from pyproj import CRS, Transformer
import rasterio
from scipy.spatial import cKDTree
from shapely import ops
from shapely.geometry import (
    LineString, MultiLineString, box, GeometryCollection,
    Polygon, MultiPolygon)

from ocsmesh.hfun.base import BaseHfun
from ocsmesh.raster import Raster, get_iter_windows
from ocsmesh.geom.shapely import PolygonGeom
from ocsmesh.features.constraint import (
        Constraint, TopoConstConstraint, TopoFuncConstraint)
from ocsmesh import utils

# supress feather warning
warnings.filterwarnings(
    'ignore', message='.*initial implementation of Parquet.*')

_logger = logging.getLogger(__name__)


def _apply_constraints(method: Callable[..., Any]) -> Callable[..., Any]:
    """Decorator for (re)applying constraint after updating hfun spec

    This function is a deocrator that takes in a callable and assumes
    the callable is a method whose first argument is an object. In the
    wrapper function, after the wrapped method is called the
    `apply_added_constraints` method is called on the object and
    then the return value of the wrapped method is returned to the
    caller.

    Parameters
    ----------
    method : callable
        Method to be wrapped so that the constraints are automatically
        re-applied after method is executed.

    Returns
    -------
    callable
        The wrapped method
    """

    def wrapped(obj, *args, **kwargs):
        rv = method(obj, *args, **kwargs)
        obj.apply_added_constraints()
        return rv
    return wrapped


class HfunInputRaster:
    """Descriptor class for holding reference to the input raster"""

    def __set__(self, obj, raster: Raster) -> None:
        if not isinstance(raster, Raster):
            raise TypeError(f'Argument raster must be of type {Raster}, not '
                            f'type {type(raster)}.')
        # init output raster file
        with ExitStack() as stack:

            src = stack.enter_context(rasterio.open(raster.tmpfile))
            if raster.chunk_size is not None:
                windows = get_iter_windows(
                    src.width, src.height, chunk_size=raster.chunk_size)
            else:
                windows = [rasterio.windows.Window(
                    0, 0, src.width, src.height)]

            meta = src.meta.copy()
            meta.update({'driver': 'GTiff', 'dtype': np.float32})
            dst = stack.enter_context(
                    obj.modifying_raster(use_src_meta=False, **meta))
            for window in windows:
                values = src.read(window=window).astype(np.float32)
                values[:] = np.finfo(np.float32).max
                dst.write(values, window=window)

        obj.__dict__['raster'] = raster
        obj._chunk_size = raster.chunk_size
        obj._overlap = raster.overlap

    def __get__(self, obj, objtype=None) -> Raster:
        return obj.__dict__['raster']


class HfunRaster(BaseHfun, Raster):
    """Raster based size function.

    Creates a raster based size function. The mesh size is specified
    at each point of the grid of the input raster, based on the
    specified criteria.

    Attributes
    ----------
    raster
    hmin
    hmax
    verbosity

    Methods
    -------
    msh_t()
        Return mesh sizes interpolated on an size-optimized
        unstructured mesh
    apply_added_constraints()
        Re-apply the existing constraint. Mostly used internally.
    apply_constraints(constraint_list)
        Apply constraint objects in the `constraint_list`.
    add_topo_bound_constraint(...)
        Add size fixed-per-point value constraint to the area
        bounded by specified bounds with expansion/contraction
        rate `rate` specified.
    add_topo_func_constraint(...)
        Add size value constraint based on function of depth/elevation
        to the area bounded by specified bounds with the expansion or
        contraction rate `rate` specified.
    add_patch(...)
        Add a region of fixed size refinement with optional expansion
        rate for points outside the region to achieve smooth size
        transition.
    add_contour(level expansion_rate target_size=None, nprocs=None)
        Add refinement based on contour lines auto-extrcted from the
        underlying raster data. The size is calculated based on the
        specified `rate`, `target_size` and the distance from the
        extracted feature line.
    add_channel(...)
        Add refinement for auto-detected narrow domain regions.
        Optionally use an expansion rate for points outside detected
        narrow regions for smooth size transition.
    add_feature(...)
        Decorated method to add size refinement based on the specified
        `expansion_rate`, `target_size`, and distance from the input
        feature lines `feature`.
    get_xy_memcache(window, dst_crs)
        Get XY grid cached onto disk. Useful for when XY needs to be
        projected to UTM so as to avoid reprojecting on every call.
    add_subtidal_flow_limiter(...)
        Add mesh size refinement based on the value as well as
        gradient of the topography within the region between
        specified by lower and upper bound on topography.
    add_constant_value(value, lower_bound=None, upper_bound=None)
        Add fixed size mesh refinement in the region specified by
        upper and lower bounds on topography.

    Notes
    -----
    Currently the implementation of this size function is such that
    when an object is created, the "size" values of this object is
    set to be maximum `np.float32`, and maximum and minimum constraints
    are not applied. After application of any refinement or constrant
    these constraints are actually applied. As a result if a raster
    based size function is created and no `add_*` or `apply_*` method
    is called, the values of `hmin` and `hmax` attributes as well as
    any value on the size function grid is meaningless.

    An important distinction that must be made is between
    **refinements** and **constraints**. These two concepts are applied
    differently when it comes to calculating the sizes. Refinements
    specification guarantees that the sizes in the specified region
    is at most equal to the specified value, however it does **not**
    have any guarantee on the minimum of sizes. In other words if
    multiple refinements are specified, then the size at everypoint
    is equal to the minimum calculated from all of those refinements.
    Contraints are applied a bit differently. Constraints ensure that
    a given condition is met at a given point. A list of specified
    constraints is created and after application of any refinements
    or other constraints, all the constraint are re-applied.
    As a result constraint can be used to ensure the size at a given
    point is not smaller that a specified value. Constraint can also
    be used similar to refinements to ensure the size is smaller
    than a specified value.

    Another important different between contraints and refinements is
    that for refinements the final value is the minimum of all, but
    constraints are applied one at a time and depending on their type
    and condition, it might result on a value between all specified
    maximums and minimums.

    Note that constraints can be conflicting, currently there's no
    automatic conflict resolutions and the constrains are applied in
    the order specified, so if applicable the last one overrides all
    else.
    """

    _raster = HfunInputRaster()

    def __init__(self,
                 raster: Raster,
                 hmin: Optional[float] = None,
                 hmax: Optional[float] = None,
                 verbosity: int = 0
                 ) -> None:
        """Initialize a raster based size function object

        Parameters
        ----------
        raster : Raster
            The input raster file. The sizes are calculated on the
            grid points of the raster's grid. The input raster is
            not modified so that its elevation data can be used for
            topo-based refinement calculations.
        hmin : float or None, default=None
            Global minimum size of mesh for the size function, if not
            specified, the calculated values during refinement and
            constraint applications are not capped off.
        hmax : float or None, default=None
            Global maximum size of mesh for the size function, if not
            specified, the calculated values during refinement and
            constraint applications are not capped off.
        verbosity : int, default=0
            The verbosity of the outputs.

        Notes
        -----
        All the points in the raster grid are set to have maximum
        `np.float32` value. The `hmin` and `hmax` arguments,
        even if provided, are not applied in during initialization.
        """

        self._xy_cache = {}
        # NOTE: unlike Raster, HfunRaster has no "path" set
        self._raster = raster
        # TODO: Store max and min as two separate constraints instead
        # of private attributes
        self._hmin = float(hmin) if hmin is not None else hmin
        self._hmax = float(hmax) if hmax is not None else hmax
        self._verbosity = int(verbosity)
        self._constraints = []


    def msh_t(
            self,
            window: Optional[rasterio.windows.Window] = None,
            marche: bool = False,
            verbosity : Optional[bool] = None
            ) -> jigsaw_msh_t:
        """Interpolates mesh size function on an unstructred mesh

        Interpolate the calculated mesh sizes from the raster grid
        onto an unstructured mesh. This mesh is generated by meshing
        the input raster using the size function values. The return
        value is in a projected CRS. If the input raster CRS is
        geographic, then a local UTM CRS is calculated and used
        for the output of this method.

        Parameters
        ----------
        window : rasterio.windows.Window or None, default=None
            If provided, a single window on raster for which the
            mesh size is to be returned.
        marche : bool, default=False
            Whether to run `marche` algorithm on the complete
            size function before calculating the unstructured mesh
            and interpolate values on it.
        verbosity : bool or None, default=None
            The verbosity of the output.

        Returns
        -------
        jigsaw_msh_t
            Size function calculated and interpolated on an
            unstructured mesh.

        Notes
        -----
        In case the underlying raster is created in windowed
        calculation mode, this method calculated the mesh for each
        window separately and then combines (no remeshing) the
        elements of all the windows.

        The output of this method needs to have length unit for
        distances (i.e. not degrees) since mesh size is specified
        in length units and the domain and size function are the
        passed to the mesh engine for cartesian meshing.

        The reason the full high-resolution size function is
        interpolated on a generated mesh it to save memory and
        have the ability to process and combine many DEMs. By doing
        more sizes are specified at points where the size needs to
        be smaller in the final mesh.

        To generate the mesh for size function interpolation, the
        raster size function (called ``hmat``) is passed to the mesh
        engine along with the bounding box of the size function as
        the meshing domain.
        """


        if window is None:
            iter_windows = list(self.iter_windows())
        else:
            iter_windows = [window]


        output_mesh = jigsaw_msh_t()
        output_mesh.ndims = +2
        output_mesh.mshID = "euclidean-mesh"
        output_mesh.crs = self.crs
        for win in iter_windows:

            hfun = jigsaw_msh_t()
            hfun.ndims = +2

            x0, y0, x1, y1 = self.get_window_bounds(win)

            utm_crs = utils.estimate_bounds_utm(
                    (x0, y0, x1, y1), self.crs)

            if utm_crs is not None:
                hfun.mshID = 'euclidean-mesh'
                # If these 3 objects (vert2, tria3, value) don't fit into
                # memroy, then the raster needs to be chunked. We need to
                # implement auto-chunking.
                start = time()
                # get bbox data
                xgrid = self.get_x(window=win)
                ygrid = np.flip(self.get_y(window=win))
                xgrid, ygrid = np.meshgrid(xgrid, ygrid)
                bottom = xgrid[0, :]
                top = xgrid[1, :]
                del xgrid
                left = ygrid[:, 0]
                right = ygrid[:, 1]
                del ygrid

                _logger.info('Building hfun.tria3...')

                dim1 = win.width
                dim2 = win.height

                tria3 = np.empty(
                    ((dim1 - 1), (dim2  - 1)),
                    dtype=jigsaw_msh_t.TRIA3_t)
                index = tria3["index"]
                helper_ary = np.ones(
                        ((dim1 - 1), (dim2  - 1)),
                        dtype=jigsaw_msh_t.INDEX_t).cumsum(1) - 1
                index[:, :, 0] = np.arange(
                        0, dim1 - 1,
                        dtype=jigsaw_msh_t.INDEX_t).reshape(dim1 - 1, 1)
                index[:, :, 0] += (helper_ary + 0) * dim1

                index[:, :, 1] = np.arange(
                        1, dim1 - 0,
                        dtype=jigsaw_msh_t.INDEX_t).reshape(dim1 - 1, 1)
                index[:, :, 1] += (helper_ary + 0) * dim1

                index[:, :, 2] = np.arange(
                        1, dim1 - 0,
                        dtype=jigsaw_msh_t.INDEX_t).reshape(dim1 - 1, 1)
                index[:, :, 2] += (helper_ary + 1) * dim1

                hfun.tria3 = tria3.ravel()
                del tria3, helper_ary
                gc.collect()
                _logger.info('Done building hfun.tria3...')

                # BUILD VERT2_t. this one comes from the memcache array
                _logger.info('Building hfun.vert2...')
                hfun.vert2 = np.empty(
                    win.width*win.height,
                    dtype=jigsaw_msh_t.VERT2_t)
                hfun.vert2['coord'] = np.array(
                    self.get_xy_memcache(win, utm_crs))
                _logger.info('Done building hfun.vert2...')

                # Build REALS_t: this one comes from hfun raster
                _logger.info('Building hfun.value...')
                hfun.value = np.array(
                    self.get_values(window=win, band=1).flatten().reshape(
                        (win.width*win.height, 1)),
                    dtype=jigsaw_msh_t.REALS_t)
                _logger.info('Done building hfun.value...')

                # Build Geom
                _logger.info('Building initial geom...')
                transformer = Transformer.from_crs(
                    self.crs, utm_crs, always_xy=True)
                bbox = [
                    *[(x, left[0]) for x in bottom],
                    *[(bottom[-1], y) for y in reversed(right)],
                    *[(x, right[-1]) for x in reversed(top)],
                    *[(bottom[0], y) for y in reversed(left)]]
                geom = PolygonGeom(
                    ops.transform(transformer.transform, Polygon(bbox)),
                    utm_crs
                ).msh_t()
                _logger.info('Building initial geom done.')
                kwargs = {'method': 'nearest'}

            else:
                _logger.info('Forming initial hmat (euclidean-grid).')
                start = time()
                hfun.mshID = 'euclidean-grid'
                hfun.xgrid = np.array(
                    np.array(self.get_x(window=win)),
                    dtype=jigsaw_msh_t.REALS_t)
                hfun.ygrid = np.array(
                    np.flip(self.get_y(window=win)),
                    dtype=jigsaw_msh_t.REALS_t)
                hfun.value = np.array(
                    np.flipud(self.get_values(window=win, band=1)),
                    dtype=jigsaw_msh_t.REALS_t)
                kwargs = {'kx': 1, 'ky': 1}  # type: ignore[dict-item]
                geom = PolygonGeom(box(x0, y1, x1, y0), self.crs).msh_t()

            _logger.info(f'Initial hfun generation took {time()-start}.')

            _logger.info('Configuring jigsaw...')

            opts = jigsaw_jig_t()

            # additional configuration options
            opts.mesh_dims = +2
            opts.hfun_scal = 'absolute'
            # no need to optimize for size function generation
            opts.optm_tria = False

            opts.hfun_hmin = np.min(hfun.value) if self.hmin is None else \
                self.hmin
            opts.hfun_hmax = np.max(hfun.value) if self.hmax is None else \
                self.hmax
            opts.verbosity = self.verbosity if verbosity is None else \
                verbosity

            # mesh of hfun window
            window_mesh = jigsaw_msh_t()
            window_mesh.mshID = 'euclidean-mesh'
            window_mesh.ndims = +2

            if marche is True:
                libsaw.marche(opts, hfun)

            libsaw.jigsaw(opts, geom, window_mesh, hfun=hfun)

            del geom
            # do post processing
            hfun.crs = utm_crs
            utils.interpolate(hfun, window_mesh, **kwargs)

            # reproject to combine with other windows
            if utm_crs is not None:
                window_mesh.crs = utm_crs
                utils.reproject(window_mesh, self.crs)


            # combine with results from previous windows
            output_mesh.tria3 = np.append(
                output_mesh.tria3,
                np.array([((idx + len(output_mesh.vert2)), tag)
                          for idx, tag in window_mesh.tria3],
                         dtype=jigsaw_msh_t.TRIA3_t),
                axis=0)
            output_mesh.vert2 = np.append(
                output_mesh.vert2,
                np.array(list(window_mesh.vert2),
                         dtype=jigsaw_msh_t.VERT2_t),
                axis=0)
            if output_mesh.value.size:
                output_mesh.value = np.append(
                    output_mesh.value,
                    np.array(list(window_mesh.value),
                             dtype=jigsaw_msh_t.REALS_t),
                    axis=0)
            else:
                output_mesh.value = np.array(
                        list(window_mesh.value),
                        dtype=jigsaw_msh_t.REALS_t)

        # NOTE: In the end we need to return in a CRS that
        # uses meters as units. UTM based on the center of
        # the bounding box of the hfun is used
        utm_crs = utils.estimate_bounds_utm(
                self.get_bbox().bounds, self.crs)
        if utm_crs is not None:
            transformer = Transformer.from_crs(
                self.crs, utm_crs, always_xy=True)
            output_mesh.vert2['coord'] = np.vstack(
                transformer.transform(
                    output_mesh.vert2['coord'][:, 0],
                    output_mesh.vert2['coord'][:, 1]
                    )).T
            output_mesh.crs = utm_crs

        return output_mesh


    def apply_added_constraints(self) -> None:
        """Apply all the added constraints

        This method is implemented for internal use. It's public
        because it needs to be called from outside the class through
        a decorator.

        Parameters
        ----------

        Returns
        -------
        None
        """

        self.apply_constraints(self._constraints)


    def apply_constraints(
            self,
            constraint_list: Iterable[Constraint]
            ) -> None:
        """Applies constraints specified by the list of contraint objects.

        Applies constraints from the provided list `constraint_list`,
        but doesn't not store them in the internal size function
        constraint list. This is mostly for internal use.

        Parameters
        ----------
        constraint_list : iterable of Constraint
            List of constraint objects to be applied to (not stored in)
            the size function.

        Returns
        -------
        None
        """

        # TODO: Validate conflicting constraints

        # Apply constraints
        with self.modifying_raster() as dst:
            iter_windows = list(self.iter_windows())
            tot = len(iter_windows)

            for i, window in enumerate(iter_windows):
                hfun_values = self.get_values(band=1, window=window)
                rast_values = self.raster.get_values(band=1, window=window)


                # Get locations
                utm_crs = utils.estimate_bounds_utm(
                        self.get_window_bounds(window), self.crs)

                if utm_crs is not None:
                    xy = self.get_xy_memcache(window, utm_crs)
                else:
                    xy = self.get_xy(window)


                # Apply custom constraints
                _logger.debug(f'Processing window {i+1}/{tot}.')
                for constraint in constraint_list:
                    hfun_values = constraint.apply(
                            rast_values, hfun_values, locations=xy)

                # Apply global constraints
                if self.hmin is not None:
                    hfun_values[hfun_values < self.hmin] = self.hmin
                if self.hmax is not None:
                    hfun_values[hfun_values > self.hmax] = self.hmax

                dst.write_band(1, hfun_values, window=window)
                del rast_values
                gc.collect()


    @_apply_constraints
    def add_topo_bound_constraint(
            self,
            value: Union[float, npt.NDArray[np.float32]],
            upper_bound: float = np.inf,
            lower_bound: float = -np.inf,
            value_type: Literal['min', 'max'] = 'min',
            rate: float = 0.01
            ) -> None:
        """Add a fixed-value or fixed-matrix constraint.

        Add a fixed-value or fixed-matrix constraint to the region
        of the size function specified by lower and upper elevation
        of the underlying DEM. Optionally a `rate` can be specified
        to relax the constraint gradually outside the bounds.

        Parameters
        ----------
        value : float or array-like
            A single fixed value or array of values to be used for
            mesh size if condition is not met based on `value_type`.
            In case of an array the dimensions must match or be
            broadcastable to the raster grid.
        upper_bound : float, default=np.inf
            Maximum elevation to cut off the region where the
            constraint needs to be applied
        lower_bound : float, default=-np.inf
            Minimum elevation to cut off the region where the
            constraint needs to be applied
        value_type : {'min', 'max'}, default='min'
            Type of contraint. If 'min', it means the mesh size
            should not be smaller than the specified `value` at each
            point.
        rate : float, default=0.01
            Rate of relaxation of constraint outside the region defined
            by `lower_bound` and `upper_bound`.

        Returns
        -------
        None

        See Also
        --------
        add_topo_func_constraint :
            Addint constraint based on function of topography
        """

        # TODO: Validate conflicting constraints, right now last one wins
        self._constraints.append(TopoConstConstraint(
            value, upper_bound, lower_bound, value_type, rate))


    @_apply_constraints
    def add_topo_func_constraint(
            self,
            func: Callable[[npt.NDArray[np.float32]], npt.NDArray[np.float32]]
                = lambda i: i / 2.0,
            upper_bound: float = np.inf,
            lower_bound: float = -np.inf,
            value_type: Literal['min', 'max'] = 'min',
            rate: float = 0.01
            ) -> None:
        """Add constraint based on a function of the topography

        Add a constraint based on the provided function `func` of
        the topography and apply to the region of the size function
        specified by lower and upper elevation of the underlying DEM.
        Optionally a `rate` can be specified to relax the constraint
        gradually outside the bounds.

        Parameters
        ----------
        func : callable
            A function to be applied on the topography to acquire the
            values to be used for mesh size if condition is not met
            based on `value_type`.
        upper_bound : float, default=np.inf
            Maximum elevation to cut off the region where the
            constraint needs to be applied
        lower_bound : float, default=-np.inf
            Minimum elevation to cut off the region where the
            constraint needs to be applied
        value_type : {'min', 'max'}, default='min'
            Type of contraint. If 'min', it means the mesh size
            should not be smaller than the value calculated from the
            specified `func` at each point.
        rate : float, default=0.01
            Rate of relaxation of constraint outside the region defined
            by `lower_bound` and `upper_bound`.

        Returns
        -------
        None

        See Also
        --------
        add_topo_bound_constraint :
            Add fixed-value or fixed-matrix constraint.
        """

        # TODO: Validate conflicting constraints, right now last one wins
        self._constraints.append(TopoFuncConstraint(
            func, upper_bound, lower_bound, value_type, rate))



    @_apply_constraints
    def add_patch(
            self,
            multipolygon: Union[MultiPolygon, Polygon],
            expansion_rate: Optional[float] = None,
            target_size: Optional[float] = None,
            nprocs: Optional[int] = None
            ) -> None:
        """Add refinement as a region of fixed size with an optional rate

        Add a refinement based on a region specified by `multipolygon`.
        The fixed `target_size` refinement can be expanded outside the
        region specified by the shape if `expansion_rate` is provided.

        Parameters
        ----------
        multipolygon : MultiPolygon or Polygon
            Shape of the region to use specified `target_size` for
            refinement.
        expansion_rate : float or None, default=None
            Optional rate to use for expanding refinement outside
            the specified shape in `multipolygon`.
        target_size : float or None, default=None
            Fixed target size of mesh to use for refinement in
            `multipolygon`
        nprocs : int or None, default=None
            Number of processors to use in parallel sections of the
            algorithm

        Returns
        -------
        None

        See Also
        --------
        add_feature :
            Add refinement for specified line string
        add_contour :
            Add refinement for auto-extracted contours
        add_channel :
            Add refinement for auto-extracted narrow regions
        add_subtidal_flow_limiter :
            Add refinement based on topograph
        add_constant_value :
            Add refinement with fixed value
        """

        # TODO: Add pool input support like add_feature for performance

        # TODO: Support other shapes - call buffer(1) on non polygons(?)
        if not isinstance(multipolygon, (Polygon, MultiPolygon)):
            raise TypeError(
                    f"Wrong type \"{type(multipolygon)}\""
                    f" for multipolygon input.")

        if isinstance(multipolygon, Polygon):
            multipolygon = MultiPolygon([multipolygon])

        # Check nprocs
        nprocs = -1 if nprocs is None else nprocs
        nprocs = cpu_count() if nprocs == -1 else nprocs
        _logger.debug(f'Using nprocs={nprocs}')


        # check target size
        target_size = self.hmin if target_size is None else target_size
        if target_size is None:
            # pylint: disable=W0101
            raise ValueError('Argument target_size must be specified if no '
                             'global hmin has been set.')
        if target_size <= 0:
            raise ValueError("Argument target_size must be greater than zero.")

        # For expansion_rate
        if expansion_rate is not None:
            exteriors = [ply.exterior for ply in multipolygon]
            interiors = [
                inter for ply in multipolygon for inter in ply.interiors]

            features = MultiLineString([*exteriors, *interiors])
            # pylint: disable=E1123, E1125
            self.add_feature(
                feature=features,
                expansion_rate=expansion_rate,
                target_size=target_size,
                nprocs=nprocs)

        with self.modifying_raster(driver='GTiff') as dst:
            iter_windows = list(self.iter_windows())
            tot = len(iter_windows)
            for i, window in enumerate(iter_windows):
                _logger.debug(f'Processing window {i+1}/{tot}.')
                # NOTE: We should NOT transform polygon, user just
                # needs to make sure input polygon has the same CRS
                # as the hfun (we don't calculate distances in this
                # method)

                _logger.info('Creating mask from shape ...')
                start = time()
                try:
                    mask, _, _ = rasterio.mask.raster_geometry_mask(
                        self.src, multipolygon,
                        all_touched=True, invert=True)
                    mask = mask[rasterio.windows.window_index(window)]

                except ValueError:
                    # If there's no overlap between the raster and
                    # shapes then it throws ValueError, instead of
                    # checking for intersection, if there's a value
                    # error we assume there's no overlap
                    _logger.debug(
                        'Polygons don\'t intersect with the raster')
                    continue
                _logger.info(
                    f'Creating mask from shape took {time()-start}.')

                values = self.get_values(window=window).copy()
                if mask.any():
                    # NOTE: Don't continue, otherwise the final
                    # destination file might end up being empty!
                    values[mask] = target_size
                if self.hmin is not None:
                    values[np.where(values < self.hmin)] = self.hmin
                if self.hmax is not None:
                    values[np.where(values > self.hmax)] = self.hmax
                values = np.minimum(self.get_values(window=window), values)

                _logger.info('Write array to file...')
                start = time()
                dst.write_band(1, values, window=window)
                _logger.info(f'Write array to file took {time()-start}.')


    @_apply_constraints
    def add_contour(
            self,
            level: Union[List[float], float],
            expansion_rate: float,
            target_size: Optional[float] = None,
            nprocs: Optional[int] = None
            ) -> None:
        """Add refinement for auto extracted contours

        Add refinement for the contour lines extracted based on
        level or levels specified by `level`. The refinement
        is relaxed with `expansion_rate` and distance from the
        extracted contour lines.

        Parameters
        ----------
        level : float or list of floats
            Level(s) at which contour lines should be extracted.
        expansion_rate : float
            Rate to use for expanding refinement with distance away
            from the extracted contours.
        target_size : float or None, default=None
            Target size to use on the extracted contours and expand
            from with distance.
        nprocs : int or None, default=None
            Number of processors to use in parallel sections of the
            algorithm

        Returns
        -------
        None

        See Also
        --------
        add_feature :
            Add refinement for specified line string
        add_patch :
            Add refinement for region specified polygon
        add_channel :
            Add refinement for auto-extracted narrow regions
        add_subtidal_flow_limiter :
            Add refinement based on topograph
        add_constant_value :
            Add refinement with fixed value

        Notes
        -----
        This method extracts contours at specified levels and
        the calls `add_feature` by passing those contour lines.
        """

        if not isinstance(level, list):
            level = [level]

        contours = []
        for _level in level:
            # pylint: disable=R1724

            _contours = self.raster.get_contour(_level)
            if isinstance(_contours, GeometryCollection):
                continue
            elif isinstance(_contours, LineString):
                contours.append(_contours)
            elif isinstance(_contours, MultiLineString):
                for _cont in _contours:
                    contours.append(_cont)

        if len(contours) == 0:
            _logger.info('No contours found!')
            return

        contours = MultiLineString(contours)

        _logger.info('Adding contours as features...')
        # pylint: disable=E1123, E1125
        self.add_feature(
                contours, expansion_rate, target_size,
                nprocs=nprocs)

    @_apply_constraints
    def add_channel(
            self,
            level: float = 0,
            width: float = 1000,
            target_size: float = 200,
            expansion_rate: Optional[float] = None,
            nprocs: Optional[int] = None,
            tolerance: Optional[float] = None
            ) -> None:
        """Add refinement for auto detected channels

        Automatically detects narrow regions in the domain and apply
        refinement size with an expanion rate (if provided) outside
        the detected area.

        Parameters
        ----------
        level : float, default=0
            High water mark at which domain is extracted for narrow
            region or channel calculations.
        width : float, default=1000
            The cut-off width for channel detection.
        target_size : float, default=200
            Target size to use on the detected channels and expand
            from with distance.
        expansion_rate : float or None, default=None
            Rate to use for expanding refinement with distance away
            from the detected channels.
        nprocs : int or None, default=None
            Number of processes to use for parallel sections of the
            workflow.
        tolerance: float or None, default=None
            Tolerance to use for simplifying the polygon extracted
            from DEM data. If `None` don't simplify.

        Returns
        -------
        None

        See Also
        --------
        add_contour :
            Add refinement for auto-extracted contours
        add_patch :
            Add refinement for region specified polygon
        add_feature :
            Add refinement for specified line string
        add_subtidal_flow_limiter :
            Add refinement based on topograph
        add_constant_value :
            Add refinement with fixed value

        Notes
        -----
        """

        channels = self.raster.get_channels(
                level=level, width=width, tolerance=tolerance)

        if channels is None:
            return

        self.add_patch(
            channels, expansion_rate, target_size, nprocs)



    @_apply_constraints
    @utils.add_pool_args
    def add_feature(
            self,
            feature: Union[LineString, MultiLineString],
            expansion_rate: float,
            target_size: Optional[float] = None,
            max_verts: int = 200,
            *, # kwarg-only comes after this
            pool: Pool,
            ) -> None:
        """Add refinement for specified linestring `feature`

        Add refinement for the specified linestrings `feature`.
        The refinement is relaxed with `expansion_rate` and distance
        from the extracted contour lines.

        Parameters
        ----------
        feature : LineString or MultiLineString
            The user-specified line strings for applying refinement on.
        expansion_rate : float
            Rate to use for expanding refinement with distance away
            from the extracted contours.
        target_size : float or None, default=None
            Target size to use on the extracted contours and expand
            from with distance.
        max_verts : int, default=200
            Number of maximum vertices in a feature line that is
            passed to a separate process in parallel section of
            the algorithm.
        pool : Pool
            Pre-created and initialized process pool to be used for
            parallel sections of the algorithm.

        Returns
        -------
        None

        See Also
        --------
        add_contour :
            Add refinement for auto-extracted contours
        add_patch :
            Add refinement for region specified polygon
        add_channel :
            Add refinement for auto-extracted narrow regions
        add_subtidal_flow_limiter :
            Add refinement based on topograph
        add_constant_value :
            Add refinement with fixed value

        Notes
        -----
        See https://outline.com/YU7nSM for an explanation
        about tree algorithms.

        Creating a local projection allows having similar area/length
        calculations as if great circle calculations was being used.

        Another useful refererence:
        https://gis.stackexchange.com/questions/214261/should-we-always-calculate-length-and-area-in-lat-lng-to-get-accurate-sizes-leng
        """

        # TODO: Consider using BallTree with haversine or Vincenty
        # metrics instead of a locally projected window.

        # TODO: Partition features if they are too "long" which results in an
        # improvement for parallel pool. E.g. if a feature is too long, 1
        # processor will be busy and the rest will be idle.

        if not isinstance(feature, (LineString, MultiLineString)):
            raise TypeError(
                f'Argument feature must be of type {LineString} or '
                f'{MultiLineString}, not type {type(feature)}.')

        if isinstance(feature, LineString):
            feature = [feature]

        elif isinstance(feature, MultiLineString):
            feature = list(feature)

        # check target size
        target_size = self.hmin if target_size is None else target_size
        if target_size is None:
            raise ValueError('Argument target_size must be specified if no '
                             'global hmin has been set.')
        if target_size <= 0:
            raise ValueError("Argument target_size must be greater than zero.")
        with self.modifying_raster(driver='GTiff') as dst:
            iter_windows = list(self.iter_windows())
            tot = len(iter_windows)
            for i, window in enumerate(iter_windows):
                _logger.debug(f'Processing window {i+1}/{tot}.')
                utm_crs = utils.estimate_bounds_utm(
                        self.get_window_bounds(window), self.crs)

                _logger.info('Repartitioning features...')
                start = time()
                res = pool.starmap(
                    utils.repartition_features,
                    [(linestring, max_verts) for linestring in feature]
                    )
                win_feature = functools.reduce(operator.iconcat, res, [])
                _logger.info(f'Repartitioning features took {time()-start}.')

                _logger.info('Resampling features on ...')
                start = time()

                # We don't want to recreate the same transformation
                # many times (it takes time) and we can't pass
                # transformation object to subtask (cinit issue)
                transformer = None
                if utm_crs is not None:
                    start2 = time()
                    transformer = Transformer.from_crs(
                        self.src.crs, utm_crs, always_xy=True)
                    _logger.info(
                            f"Transform creation took {time() - start2:f}")
                    start2 = time()
                    win_feature = [
                        ops.transform(transformer.transform, linestring)
                        for linestring in win_feature]
                    _logger.info(
                            f"Transform apply took {time() - start2:f}")

                transformed_features = pool.starmap(
                    utils.transform_linestring,
                    [(linestring, target_size) for linestring in win_feature]
                )
                _logger.info(f'Resampling features took {time()-start}.')
                _logger.info('Concatenating points...')
                start = time()
                points = []
                for geom in transformed_features:
                    if isinstance(geom, LineString):
                        points.extend(geom.coords)
                    elif isinstance(geom, MultiLineString):
                        for linestring in geom:
                            points.extend(linestring.coords)
                _logger.info(f'Point concatenation took {time()-start}.')

                _logger.info('Generating KDTree...')
                start = time()
                tree = cKDTree(np.array(points))
                _logger.info(f'Generating KDTree took {time()-start}.')
                if utm_crs is not None:
                    xy = self.get_xy_memcache(window, utm_crs)
                else:
                    xy = self.get_xy(window)

                _logger.info(f'Transforming points took {time()-start}.')
                _logger.info('Querying KDTree...')
                start = time()
                if self.hmax:
                    r = (self.hmax - target_size) / (expansion_rate * target_size)
                    near_dists, neighbors = tree.query(
                        xy, workers=pool._processes, distance_upper_bound=r)
                    distances = r * np.ones(len(xy))
                    mask = np.logical_not(np.isinf(near_dists))
                    distances[mask] = near_dists[mask]
                else:
                    distances, _ = tree.query(xy, workers=pool._processes)
                _logger.info(f'Querying KDTree took {time()-start}.')
                values = expansion_rate*target_size*distances + target_size
                values = values.reshape(window.height, window.width).astype(
                    self.dtype(1))
                if self.hmin is not None:
                    values[np.where(values < self.hmin)] = self.hmin
                if self.hmax is not None:
                    values[np.where(values > self.hmax)] = self.hmax
                values = np.minimum(self.get_values(window=window), values)
                _logger.info('Write array to file...')
                start = time()
                dst.write_band(1, values, window=window)
                _logger.info(f'Write array to file took {time()-start}.')

    def get_xy_memcache(
            self,
            window : rasterio.windows.Window,
            dst_crs: Union[CRS, str]
            ) -> npt.NDArray[float]:
        """Get the transformed locations of raster points.

        Get the locations of raster points in the `dst_crs` CRS.
        This method caches these transformed values for fast retrieval
        upon multiple calls.

        Parameters
        ----------
        window : rasterio.windows.Window
            The raster window for querying location data.
        dst_crs : CRS or str
            The destination CRS for the raster points locations.

        Returns
        -------
        np.ndarray
            Locations of raster points after projecting to `dst_crs`

        See Also
        --------
        get_xy :
            Get the locations of raster points from the raster file.
        """

        tmpfile = self._xy_cache.get(f'{window}{dst_crs}')
        if tmpfile is None:
            _logger.info('Transform points to local CRS...')
            transformer = Transformer.from_crs(
                self.src.crs, dst_crs, always_xy=True)
            # pylint: disable=R1732
            tmpfile = tempfile.NamedTemporaryFile()
            xy = self.get_xy(window)
            fp = np.memmap(tmpfile, dtype='float32', mode='w+', shape=xy.shape)
            fp[:] = np.vstack(
                transformer.transform(xy[:, 0], xy[:, 1])).T
            _logger.info('Saving values to memcache...')
            fp.flush()
            _logger.info('Done!')
            self._xy_cache[f'{window}{dst_crs}'] = tmpfile
            return fp[:]

        _logger.info('Loading values from memcache...')
        return np.memmap(tmpfile, dtype='float32', mode='r',
                         shape=((window.width*window.height), 2))[:]

    @_apply_constraints
    def add_subtidal_flow_limiter(
            self,
            hmin: Optional[float] = None,
            hmax: Optional[float] = None,
            lower_bound: Optional[float] = None,
            upper_bound: Optional[float] = None
            ) -> None:
        """Add mesh refinement based on topography.

        Calculates a pre-defined function of topography to use
        as values for refinement. The function values are cut off
        using `hmin` and `hmax` and is applied to the region bounded
        by `lower_bound` and `upper_bound`.

        Parameters
        ----------
        hmin : float or None, default=None
            Minimum mesh size in the refinement
        hmax : float or None, default=None
            Maximum mesh size in the refinement
        lower_bound : float or None, default=None
            Lower limit of the cut-off elevation for region to apply
            the fixed `value`.
        upper_bound : float or None, default=None
            Higher limit of the cut-off elevation for region to apply
            the fixed `value`.

        Returns
        -------
        None

        See Also
        --------
        add_feature :
            Add refinement for specified line string
        add_contour :
            Add refinement for auto-extracted contours
        add_patch :
            Add refinement for region specified polygon
        add_channel :
            Add refinement for auto-extracted narrow regions
        add_constant_value :
            Add refinement with fixed value

        Notes
        -----
        The size refinement value is calculated from the topography
        using:

        .. math::
            h = {1 \\over 3} \\times {{|z|} \\over {\\left\\| \\grad z \\right\\|}}

        where :math:`z` is the elevation and :math:`h` is the value of
        the mesh size. This refinement is not applied wherever the
        magnitude of the gradient of topography is equal to zero.
        """

        hmin = float(hmin) if hmin is not None else hmin
        hmax = float(hmax) if hmax is not None else hmax

        with self.modifying_raster() as dst:

            iter_windows = list(self.iter_windows())
            tot = len(iter_windows)

            for i, window in enumerate(iter_windows):

                _logger.debug(f'Processing window {i+1}/{tot}.')

                x0, y0, x1, y1 = self.get_window_bounds(window)
                utm_crs = utils.estimate_bounds_utm(
                        (x0, y0, x1, y1), self.crs)
                if utm_crs is not None:
                    transformer = Transformer.from_crs(
                            self.crs, utm_crs, always_xy=True)
                    (x0, x1), (y0, y1) = transformer.transform(
                            [x0, x1], [y0, y1])
                    dx = np.diff(np.linspace(x0, x1, window.width))[0]
                    dy = np.diff(np.linspace(y0, y1, window.height))[0]
                else:
                    dx = self.dx
                    dy = self.dy
                topobathy = self.raster.get_values(band=1, window=window)
                dx, dy = np.gradient(topobathy, dx, dy)
                with warnings.catch_warnings():
                    # in case self._src.values is a masked array
                    warnings.simplefilter("ignore", category=RuntimeWarning)
                    dh = np.sqrt(dx**2 + dy**2)
                dh = np.ma.masked_equal(dh, 0.)
                hfun_values = np.abs((1./3.)*(topobathy/dh))
                # values = values.filled(np.max(values))

                if upper_bound is not None:
                    idxs = np.where(topobathy > upper_bound)
                    hfun_values[idxs] = self.get_values(
                        band=1, window=window)[idxs]
                if lower_bound is not None:
                    idxs = np.where(topobathy < lower_bound)
                    hfun_values[idxs] = self.get_values(
                        band=1, window=window)[idxs]

                if hmin is not None:
                    hfun_values[np.where(hfun_values < hmin)] = hmin

                if hmax is not None:
                    hfun_values[np.where(hfun_values > hmax)] = hmax

                if self._hmin is not None:
                    hfun_values[np.where(hfun_values < self._hmin)] = self._hmin
                if self._hmax is not None:
                    hfun_values[np.where(hfun_values > self._hmax)] = self._hmax

                hfun_values = np.minimum(
                    self.get_values(band=1, window=window),
                    hfun_values).astype(
                    self.dtype(1))
                dst.write_band(1, hfun_values, window=window)

    @_apply_constraints
    def add_constant_value(
            self,
            value: float,
            lower_bound: Optional[float] = None,
            upper_bound: Optional[float] = None
            ) -> None:
        """Add refinement of fixed value in the region specified by bounds.

        Apply fixed value mesh size refinement to the region specified
        by bounds (if provided) or the whole domain.

        Parameters
        ----------
        value : float
            Fixed value to use for refinement size
        lower_bound : float or None, default=None
            Lower limit of the cut-off elevation for region to apply
            the fixed `value`.
        upper_bound : float or None, default=None
            Higher limit of the cut-off elevation for region to apply
            the fixed `value`.

        Returns
        -------
        None

        See Also
        --------
        add_feature :
            Add refinement for specified line string
        add_contour :
            Add refinement for auto-extracted contours
        add_patch :
            Add refinement for region specified polygon
        add_channel :
            Add refinement for auto-extracted narrow regions
        add_subtidal_flow_limiter :
            Add refinement based on topograph
        """

        lower_bound = -float('inf') if lower_bound is None \
            else float(lower_bound)
        upper_bound = float('inf') if upper_bound is None \
            else float(upper_bound)

        with self.modifying_raster() as dst:

            iter_windows = list(self.iter_windows())
            tot = len(iter_windows)

            for i, window in enumerate(iter_windows):

                _logger.debug(f'Processing window {i+1}/{tot}.')
                hfun_values = self.get_values(band=1, window=window)
                rast_values = self.raster.get_values(band=1, window=window)
                hfun_values[np.where(np.logical_and(
                    rast_values > lower_bound,
                    rast_values < upper_bound))] = value
                hfun_values = np.minimum(
                    self.get_values(band=1, window=window),
                    hfun_values.astype(self.dtype(1)))
                dst.write_band(1, hfun_values, window=window)
                del rast_values
                gc.collect()

    @property
    def raster(self):
        """Read-only attribute to reference to the input raster"""

        return self._raster

    @property
    def hmin(self):
        """Read-only attribute for the minimum mesh size constraint"""

        return self._hmin

    @property
    def hmax(self):
        """Read-only attribute for the maximum mesh size constraint"""

        return self._hmax

    @property
    def verbosity(self):
        """Modifiable attribute for the verbosity of the output"""

        return self._verbosity

    @verbosity.setter
    def verbosity(self, verbosity: int):
        self._verbosity = verbosity


def transform_point(
        x: npt.NDArray[float],
        y: npt.NDArray[float],
        src_crs: Union[CRS, str],
        utm_crs: Union[CRS, str],
        ) -> Tuple[npt.NDArray[float], npt.NDArray[float]]:
    """Transform input locations to the destination CRS `utm_crs`

    Parameters
    ----------
    x : npt.NDArray[float]
        Vector of x locations.
    y : npt.NDArray[float]
        Vector of y locations.
    src_crs : Union[CRS, str]
        Source CRS for location transformation.
    utm_crs : Union[CRS, str]
        Destination CRS for location transformation.

    Returns
    -------
    Tuple[npt.NDArray[float], npt.NDArray[float]]
        Tuplpe of transformed x and y arrays.
    """
    transformer = Transformer.from_crs(src_crs, utm_crs, always_xy=True)
    return transformer.transform(x, y)


def transform_polygon(
    polygon: Polygon,
    src_crs: CRS = None,
    utm_crs: CRS = None
    ) -> Polygon:
    """Transform the input polygon to the destination CRS `utm_crs`

    Parameters
    ----------
    polygon : Polygon
        Input polygon to be transformed from `src_crs` to `utm_crs`.
    src_crs : Union[CRS, str]
        Source CRS for location transformation.
    utm_crs : Union[CRS, str]
        Destination CRS for location transformation.

    Returns
    -------
    Polygon
        Transformed polygon in the destination `utm_crs`.
    """

    if utm_crs is not None:
        transformer = Transformer.from_crs(
            src_crs, utm_crs, always_xy=True)

        polygon = ops.transform(
                transformer.transform, polygon)
    return polygon



import gc
import logging
from multiprocessing import Pool, Lock, cpu_count
import os
import pathlib
import tempfile
import warnings
from typing import Union, Sequence, Tuple, List

import geopandas as gpd
import numpy as np
from pyproj import CRS, Transformer
from shapely import ops
from shapely.geometry import box, Polygon, MultiPolygon, LinearRing
from shapely.validation import explain_validity

from jigsawpy import jigsaw_msh_t, savemsh, savevtk

from ocsmesh.raster import Raster
from ocsmesh.mesh.mesh import Mesh


_logger = logging.getLogger(__name__)

class GeomCombine:

    _base_mesh_lock = Lock()
    def __init__(
            self,
            dem_files: Union[None, Sequence[Union[str, os.PathLike]]],
            out_file: Union[str, os.PathLike],
            out_format: str = "shapefile",
            mesh_file: Union[str, os.PathLike, None] = None,
            mesh_multipolygon: Union[MultiPolygon, Polygon] = None,
            ignore_mesh_final_boundary : bool = False,
            zmin: Union[float, None] = None,
            zmax: Union[float, None] = None,
            chunk_size: Union[int, None] = None,
            overlap: Union[int, None] = None,
            nprocs: int = -1,
            out_crs: Union[str, CRS] = "EPSG:4326",
            base_crs: Union[str, CRS] = None):

        self._calc_crs = None
        self._base_exterior = None

        nprocs = cpu_count() if nprocs == -1 else nprocs
        dem_files = [] if dem_files is None else dem_files

        self._operation_info = dict(
            dem_files=dem_files,
            out_file=out_file,
            out_format=out_format,
            mesh_file=mesh_file,
            mesh_mp_in=mesh_multipolygon,
            ignore_mesh=ignore_mesh_final_boundary,
            zmin=zmin,
            zmax=zmax,
            chunk_size=chunk_size,
            overlap=overlap,
            nprocs=nprocs,
            out_crs=out_crs,
            base_crs=base_crs)

    def run(self):

        dem_files = self._operation_info['dem_files']
        out_file = self._operation_info['out_file']
        out_format = self._operation_info['out_format']
        mesh_file = self._operation_info['mesh_file']
        mesh_mp_in = self._operation_info['mesh_mp_in']
        ignore_mesh = self._operation_info['ignore_mesh']
        zmin = self._operation_info['zmin']
        zmax = self._operation_info['zmax']
        chunk_size = self._operation_info['chunk_size']
        overlap = self._operation_info['overlap']
        nprocs = self._operation_info['nprocs']
        out_crs = self._operation_info['out_crs']
        base_crs = self._operation_info['base_crs']

        out_dir = pathlib.Path(out_file).parent
        out_dir.mkdir(exist_ok=True, parents=True)

        # Warping takes time; to optimize, only warp rasters
        # during calculation of polygons if needed. Otherwise
        # only warp polygon before writing to file
        if isinstance(out_crs, str):
            out_crs = CRS.from_user_input(out_crs)
        if isinstance(base_crs, str):
            base_crs = CRS.from_user_input(base_crs)
        all_crs = set(Raster(dem).crs for dem in dem_files)
        self._calc_crs = out_crs
        if len(all_crs) == 1:
            self._calc_crs = list(all_crs)[0]
            _logger.info(
                f"All DEMs have the same CRS:"
                f" {self._calc_crs.to_string()}")

        base_mult_poly = None
        if mesh_mp_in:
            # Assumption: If base_mult_poly is provided, it's in
            # base_crs if not None, else in out_crs
            base_mult_poly = self._get_valid_multipolygon(mesh_mp_in)
            if base_crs is None:
                base_crs = out_crs
            if not base_crs.equals(self._calc_crs):
                _logger.info("Reprojecting base polygon...")
                transformer = Transformer.from_crs(
                    base_crs, self._calc_crs, always_xy=True)
                base_mult_poly = ops.transform(
                        transformer.transform, base_mult_poly)

        elif mesh_file and pathlib.Path(mesh_file).is_file():
            _logger.info("Creating mesh object from file...")
            base_mesh = Mesh.open(mesh_file, crs=base_crs)
            mesh_crs = base_mesh.crs
            # Assumption: If mesh_crs is not defined, mesh is in
            # base_crs if not None, else inout_crs
            if base_crs is None:
                if mesh_crs:
                    base_crs = mesh_crs
                else:
                    base_crs = out_crs
            if not self._calc_crs.equals(base_crs):
                _logger.info("Reprojecting base mesh...")
                transformer = Transformer.from_crs(
                    base_crs, self._calc_crs, always_xy=True)
                xy = base_mesh.coord
                xy = np.vstack(
                    transformer.transform(xy[:, 0], xy[:, 1])).T
                base_mesh.coord[:] = xy

            _logger.info("Done")

            _logger.info("Getting mesh hull polygons...")
            base_mult_poly = base_mesh.hull.multipolygon()
            _logger.info("Done")

            base_mult_poly = self._get_valid_multipolygon(base_mult_poly)


        if base_mult_poly:
            # NOTE: This needs to happen once and before any
            # modification to basemesh happens (due to overlap
            # w/ DEM, etc.). Exterior of base mesh is used for
            # raster clipping
            #
            # TODO: Add buffer for base mesh exterior in case there
            # was erosion and we want to make sure new DEMs futher
            # inland are considered (?)
            self._base_exterior = MultiPolygon(
                    list(ops.polygonize(
                        [poly.exterior for poly in base_mult_poly.geoms])))


        z_info = {}
        if zmin is not None:
            z_info['zmin'] = zmin
        if zmax is not None:
            z_info['zmax'] = zmax

        poly_files_coll = []
        _logger.info(f"Number of processes: {nprocs}")
        with tempfile.TemporaryDirectory(dir=out_dir) as temp_dir, \
                tempfile.NamedTemporaryFile() as base_file:

            if base_mult_poly:
                base_mesh_path = base_file.name
                self._multipolygon_to_disk(
                    base_mesh_path, base_mult_poly, fix=False)
            else:
                base_mesh_path = None
            base_mult_poly = None


            _logger.info("Processing DEM priorities ...")
            # Process priority: priority is based on the order,
            # the last input has the highest priority
            # (i.e. lowest priority number)
            priorities = list((range(len(dem_files))))[::-1]
            # TODO: Needs some code refinement for bbox issue
#            priority_args = []
#            for priority, dem_file in zip(priorities, dem_files):
#                priority_args.append(
#                    (priority, temp_dir, dem_file, chunk_size, overlap))
#
#            with Pool(processes=nprocs) as p:
#                p.starmap(self._process_priority, priority_args)
#            p.join()

            _logger.info("Processing DEM contours ...")
            # Process contours
            if nprocs > 1:
                parallel_args = []
                for priority, dem_file in zip(priorities, dem_files):
                    parallel_args.append(
                        (base_mesh_path, temp_dir,
                         priority, dem_file,
                         z_info, chunk_size, overlap))
                with Pool(processes=nprocs) as p:
                    poly_files_coll.extend(
                        p.starmap(
                            self._parallel_get_polygon_worker,
                            parallel_args))
                p.join()
            else:
                poly_files_coll.extend(
                    self._serial_get_polygon(
                        base_mesh_path, temp_dir,
                        priorities, dem_files,
                        z_info, chunk_size, overlap))


            _logger.info("Generating final boundary polygon...")
            # If a DEM doesn't intersect domain None will
            # be returned by worker
            poly_files_coll = [i for i in poly_files_coll if i]
            if base_mesh_path is not None and not ignore_mesh:
                poly_files_coll.append(base_mesh_path)

            rasters_gdf = gpd.GeoDataFrame(
                    columns=['geometry'],
                    crs=self._calc_crs
                )
            for feather_f in poly_files_coll:
                rasters_gdf = rasters_gdf.append(
                    gpd.GeoDataFrame(
                        {'geometry': self._read_multipolygon(
                                                feather_f)
                        },
                        crs=self._calc_crs
                        ),
                    ignore_index=True)


            # The assumption is this returns polygon or multipolygon
            fin_mult_poly = rasters_gdf.unary_union
            _logger.info("Done")


        # If DEM is not inside input base polygon, the end results
        # is None
        if fin_mult_poly:
            # Get a clean multipolygon to write to output
            # Is this necessary? It can be expensive if geom is not valid
            fin_mult_poly = self._get_valid_multipolygon(fin_mult_poly)

            self._write_to_file(
                    out_format, out_file, fin_mult_poly, out_crs)

        self._base_exterior = None

    def _get_valid_multipolygon(
            self,
            polygon: Union[Polygon, MultiPolygon]
            ) -> MultiPolygon:

        if not polygon.is_valid:
            polygon = ops.unary_union(polygon)

            if not polygon.is_valid:
                polygon = polygon.buffer(0)

            if not polygon.is_valid:
                raise ValueError(explain_validity(polygon))

        if isinstance(polygon, Polygon):
            polygon = MultiPolygon([polygon])

        return polygon


    def _multipolygon_to_disk(
            self,
            path: Union[str, os.PathLike],
            multipolygon: MultiPolygon,
            fix: bool = True):

        if fix:
            multipolygon = self._get_valid_multipolygon(
                    multipolygon)

        if isinstance(multipolygon, Polygon):
            # In case fix is not True, we need to make sure it's
            # a multipolygon instead of polygon for dataframe creation
            multipolygon = MultiPolygon([multipolygon])

        gpd.GeoDataFrame({'geometry': multipolygon}).to_feather(path)


    def _read_multipolygon(
            self,
            path: Union[str, os.PathLike],
            fix: bool = True
            ) -> MultiPolygon:

        multipolygon = MultiPolygon(
                list(gpd.read_feather(path).geometry))

        if fix:
            multipolygon = self._get_valid_multipolygon(
                    multipolygon)

        return multipolygon

    def _read_to_geodf(
            self,
            path: Union[str, os.PathLike],
            ) -> gpd.GeoDataFrame:

        gdf = gpd.read_feather(path)

        return gdf


    def _process_priority(
            self,
            priority: int,
            temp_dir: Union[str, os.PathLike],
            dem_path: Union[str, os.PathLike],
            chunk_size: Union[int, None] = None,
            overlap: Union[int, None] = None):

        rast = Raster(
                dem_path,
                chunk_size=chunk_size,
                overlap=overlap)
        # Can cause issue with bbox(?)
        if not self._calc_crs.equals(rast.crs):
            rast.warp(dst_crs=self._calc_crs)

        pri_dt_path = (
            pathlib.Path(temp_dir) / f'dem_priority_{priority}.feather')

        pri_mult_poly = MultiPolygon([box(*rast.src.bounds)])

        self._multipolygon_to_disk(
                pri_dt_path, pri_mult_poly)


    def _serial_get_polygon(
            self,
            base_mesh_path: Union[str, os.PathLike, None],
            temp_dir: Union[str, os.PathLike],
            priorities: Sequence[int],
            dem_files: Sequence[Union[str, os.PathLike]],
            z_info: dict = None,
            chunk_size: Union[int, None] = None,
            overlap: Union[int, None] = None):

        if z_info is None:
            z_info = {}

        _logger.info("Getting DEM info")
        poly_coll = []
        for priority, dem_path in zip(priorities, dem_files):
            _logger.info(f"Processing {dem_path} ...")
            if not pathlib.Path(dem_path).is_file():
                warnings.warn(f"File {dem_path} not found!")
                _logger.debug(f"File {dem_path} not found!")
                continue

            # Calculate Polygon
            _logger.info("Loading raster from file...")
            rast = Raster(
                    dem_path,
                    chunk_size=chunk_size,
                    overlap=overlap)
            # Can cause issue with bbox(?)
            if not self._calc_crs.equals(rast.crs):
                rast.warp(dst_crs=self._calc_crs)

            _logger.info("Clipping to basemesh size if needed...")
            rast_box = box(*rast.src.bounds)
            if base_mesh_path is not None:
                # NOTE: We use the exterior from the earlier calc
                if self._base_exterior and not rast_box.within(self._base_exterior):
                    if not rast_box.intersects(self._base_exterior):
                        _logger.info(
                            f"{dem_path} is ignored due to base mesh...")
                        continue

                    _logger.info(
                        f"{dem_path} needs clipping by base mesh...")
                    rast.clip(self._base_exterior)
                    rast_box = box(*rast.src.bounds)

            # Processing raster
            _logger.info("Creating geom from raster...")

            _logger.info("Getting polygons from geom...")
            geom_mult_poly = rast.get_multipolygon(**z_info)
            geom_mult_poly = self._get_valid_multipolygon(
                    geom_mult_poly)

            if base_mesh_path is not None:
                _logger.info("Subtract DEM bounds from base mesh polygons...")
                self._base_mesh_lock.acquire()
                try:
                    # Get a valid multipolygon from disk
                    base_mult_poly = self._read_multipolygon(
                            base_mesh_path)

                    # Get valid multipolygon after operation and write
                    base_mult_poly = base_mult_poly.difference(
                            rast_box)
                    self._multipolygon_to_disk(
                            base_mesh_path, base_mult_poly)

                finally:
                    self._base_mesh_lock.release()

            # TODO: Needs some code refinement due to bbox
            # Processing DEM priority
#            priority_geodf = gpd.GeoDataFrame(
#                    columns=['geometry'],
#                    crs=self._calc_crs)
#            for p in range(priority):
#                higher_pri_path = (
#                    pathlib.Path(temp_dir) / f'dem_priority_{p}.feather')
#
#                if higher_pri_path.is_file():
#                    priority_geodf = priority_geodf.append(
#                             self._read_to_geodf(higher_pri_path))
#
#            if len(priority_geodf):
#                op_res = priority_geodf.unary_union
#                pri_mult_poly = MultiPolygon()
#                if isinstance(op_res, MultiPolygon):
#                    pri_mult_poly = op_res
#                else:
#                    pri_mult_poly = MultiPolygon([op_res])
#
#
#                if rast_box.within(pri_mult_poly):
#                    _logger.info(
#                        f"{dem_path} is ignored due to priority...")
#                    continue
#
#                if rast_box.intersects(pri_mult_poly):
#                    _logger.info(
#                        f"{dem_path} needs clipping by priority...")
#
#                    # Clipping raster can cause problem at
#                    # boundaries due to difference in pixel size
#                    # between high and low resolution rasters
#                    # so instead we operate on extracted polygons
#                    geom_mult_poly = geom_mult_poly.difference(
#                            pri_mult_poly)


            # Write geometry multipolygon to disk
            temp_path = (
                    pathlib.Path(temp_dir)
                    / f'{pathlib.Path(dem_path).name}.feather')

            try:
                self._multipolygon_to_disk(temp_path, geom_mult_poly)
                poly_coll.append(temp_path)
            except:
                warnings.warn(f"Error writing {temp_path} to disk")

            # Multipolygon takes a lot of memory
            del geom_mult_poly
            gc.collect(2)

        return poly_coll


    def _parallel_get_polygon_worker(
            self,
            base_mesh_path: Union[str, os.PathLike, None],
            temp_dir: Union[str, os.PathLike],
            priority: int,
            dem_file: Union[str, os.PathLike],
            z_info: dict = None,
            chunk_size: Union[int, None] = None,
            overlap: Union[int, None] = None):

        if z_info is None:
            z_info = {}

        poly_coll_files = self._serial_get_polygon(
            base_mesh_path, temp_dir, [priority], [dem_file],
            z_info, chunk_size, overlap)

        # Only one item passed to serial code at most
        return poly_coll_files[0] if poly_coll_files else None


    def _linearring_to_vert_edge(
            self,
            coords: List[Tuple[float, float]],
            edges: List[Tuple[int, int]],
            lin_ring: LinearRing):

        '''From shapely LinearRing get coords and edges'''

        # NOTE: This function mutates coords and edges

        # TODO: Move to utils?
        idx_b = len(coords)
        coords.extend(coord for coord in lin_ring.coords)
        # Last coord is the same as first in a ring
        coords.pop()
        idx_e = len(coords) - 1
        n_idx = len(coords)

        edges.extend([
            (i, (i + 1) % n_idx + idx_b * ((i + 1) // n_idx))
            for i in range(idx_b, idx_e + 1)])


    def _write_to_file(
            self, out_format, out_file, multi_polygon, crs):

        _logger.info(f"Writing for file ({out_format}) ...")

        # TODO: Check for correct extension on out_file
        if out_format == "shapefile":
            gdf = gpd.GeoDataFrame(
                    {'geometry': multi_polygon},
                    crs=self._calc_crs
                    )
            if not crs.equals(self._calc_crs):
                _logger.info(
                    f"Project from {self._calc_crs.to_string()} to"
                    f" {crs.to_string()} ...")
                gdf = gdf.to_crs(crs)
            gdf.to_file(out_file)

        elif out_format == "feather":
            gdf = gpd.GeoDataFrame(
                    {'geometry': multi_polygon},
                    crs=self._calc_crs
                    )
            if not crs.equals(self._calc_crs):
                _logger.info(
                    f"Project from {self._calc_crs.to_string()} to"
                    f" {crs.to_string()} ...")
                gdf = gdf.to_crs(crs)
            gdf.to_feather(out_file)

        elif out_format in ("jigsaw", "vtk"):

            if not crs.equals(self._calc_crs):
                _logger.info(
                    f"Project from {self._calc_crs.to_string()} to"
                    f" {crs.to_string()} ...")
                transformer = Transformer.from_crs(
                    self._calc_crs, crs, always_xy=True)
                multi_polygon = ops.transform(
                        transformer.transform, multi_polygon)

            msh = jigsaw_msh_t()
            msh.ndims = +2
            msh.mshID = 'euclidean-mesh'

            coords = []
            edges = []
            for polygon in multi_polygon:
                self._linearring_to_vert_edge(
                        coords, edges, polygon.exterior)
                for interior in polygon.interiors:
                    self._linearring_to_vert_edge(
                            coords, edges, interior)

            msh.vert2 = np.array(
                [(i, 0) for i in coords],
                dtype=jigsaw_msh_t.VERT2_t)
            msh.edge2 = np.array(
                [(i, 0) for i in edges],
                dtype=jigsaw_msh_t.EDGE2_t)


            if out_format == "jigsaw":
                savemsh(out_file, msh)

            elif out_format == "vtk":
                savevtk(out_file, msh)

        else:
            raise NotImplementedError(f"Output type {out_format} is not supported")

        _logger.info("Done")



#!/usr/bin/env python
"""
CLI interface for interpolating rasters into a mesh.
"""
import sys
import argparse
import pathlib
import shutil
import tempfile
from functools import lru_cache
from multiprocessing import Pool

import numpy as np
import fiona
import requests
from tqdm import tqdm
from matplotlib.path import Path  # type: ignore[import]
from scipy.interpolate import RectBivariateSpline  # , griddata
from geoalchemy2.shape import from_shape
from shapely.geometry import box
from shapely.ops import transform
from pyproj import CRS, Transformer
from ocsmesh import Mesh, Raster, db


def _mesh_interpolate_worker(args):
    coords, raster, crs, chunk_size = args
    raster = Raster(raster)
    raster.warp(crs)
    results = []
    for window in raster.iter_windows(chunk_size=chunk_size, overlap=2):
        xi = raster.get_x(window)
        yi = raster.get_y(window)
        zi = raster.get_values(window=window)
        f = RectBivariateSpline(
            xi, np.flip(yi), np.fliplr(zi).T,
            bbox=[
                np.min(xi),
                np.max(xi),
                np.min(yi),
                np.max(yi)],
            kx=3,
            ky=3,
            s=0
        )
        idxs = np.where(
            np.logical_and(
                np.logical_and(
                    np.min(xi) < coords[:, 0],
                    np.max(xi) > coords[:, 0]),
                np.logical_and(
                    np.min(yi) < coords[:, 1],
                    np.max(yi) > coords[:, 1])))[0]

        values = f.ev(coords[idxs, 0], coords[idxs, 1])
        results.append((idxs, values))
    return results


class Interp:

    def __init__(self, args):
        self._args = args

    def main(self):

        if self._args.nprocs is None:
            self._main_serial()
        else:
            self._main_parallel()

        # write output to file
        self.mesh.write(
            self._args.output_mesh_path,
            overwrite=self._args.overwrite,
            fmt='gr3'
            )
        self.mesh.write(
            self._args.output_mesh_path + '.2dm',
            overwrite=self._args.overwrite,
            fmt='2dm'
            )

    def _main_serial(self):
        if self._args.use_anti_aliasing is True:
            self._initial_values = self.mesh.values.copy()

        for tile in (
                tqdm(self._rasters) if self._args.verbose else self._rasters):
            raster = Raster(tile)
            self.mesh.interpolate(raster)

            if self._args.use_anti_aliasing is True:
                self._resolve_aliasing(raster)

    def _main_parallel(self):
        # https://pythonspeed.com/articles/python-multiprocessing/
        # https://stackoverflow.com/questions/29864707/python-multiprocessing-for-matplotlib-griddata
        # mesh_values = self.mesh.values.copy()

        chunk_size = self._args.chunk_size
        tmpfile = tempfile.NamedTemporaryFile()
        fp = np.memmap(
            tmpfile.name,
            dtype='float32',
            mode='w+',
            shape=self.mesh.coords.shape
            )
        fp[:] = self.mesh.coords[:]
        with Pool(processes=self._args.nprocs) as pool:
            if self._args.verbose:
                res = list(tqdm(pool.imap(
                    _mesh_interpolate_worker,
                    [(fp, raster, self.mesh.crs, chunk_size)
                     for raster in self._rasters]
                    ), total=len(self._rasters)))
            else:
                res = pool.map(
                    _mesh_interpolate_worker,
                    [(fp, raster, self.mesh.crs, chunk_size)
                     for raster in self._rasters]
                    )
        del tmpfile
        output = np.full(self.mesh.values.shape, np.nan)
        for _ in res:
            for idxs, values in _:
                for i, idx in enumerate(idxs):
                    output[idx] = np.nanmean([output[idx], values[i]])
        idxs = np.where(np.isnan(output))
        output[idxs] = self.mesh.values[idxs]
        self.mesh._values = output

    @property
    @lru_cache(maxsize=None)
    def mesh(self):
        return Mesh.open(
            pathlib.Path(self._args.input_mesh_path),
            CRS.from_user_input(self._args.crs)
            )

    def _resolve_aliasing(self, raster):
        xy = self.mesh.get_xy(crs=raster.crs)
        rbbox = raster.bbox
        idxs = np.where(np.logical_and(
            np.logical_and(xy[:, 0] >= rbbox.xmin, xy[:, 0] <= rbbox.xmax),
            np.logical_and(xy[:, 1] >= rbbox.ymin, xy[:, 1] <= rbbox.ymax)
            ))

        zero_cross = np.sign(
            self.mesh._values[idxs]) - np.sign(self._initial_values[idxs])
        if np.any(zero_cross == 0.):
            _idxs = np.where(zero_cross != 0.)
            if self._args.anti_aliasing_method == "reuse":
                self.mesh._values[idxs][_idxs] = self._initial_values[idxs][_idxs]

            elif self._args.anti_aliasing_method == "fv":
                all_rings = []
                for rings in self.mesh.index_ring_collection.values():
                    all_rings.extend([rings['exterior'], *rings['interiors']])

                for i, idx in enumerate(idxs):
                    if zero_cross[i] == 0:
                        vertices = []
                        x0, y0 = xy[idx, 0], xy[idx, 1]

                        # include source point if it's a boundary node
                        for ring in all_rings:
                            if idx in ring:
                                vertices.append((x0, y0))
                                break

                        # find midpoint between neighbors
                        for neigh in self.mesh.node_neighbors[idx]:
                            vertices.append(
                                ((x0 + xy[neigh, 0])/2, (y0 + xy[neigh, 1])/2))

                        # compute centroids of neighbors
                        elements = self.mesh.triangulation.triangles[
                            np.where(
                                np.any(
                                    np.isin(
                                        self.mesh.triangulation.triangles,
                                        idx
                                        ),
                                    axis=1
                                    )
                                )
                            ]
                        cx = np.sum(self.mesh.x[elements], axis=1) / 3
                        cy = np.sum(self.mesh.y[elements], axis=1) / 3
                        for j in range(elements.shape[0]):
                            vertices.append((cx[j], cy[j]))
                        # finally make sure the vertices are ordered.
                        vertices = list(polygon_sort(vertices))
                        vertices.append(vertices[0])
                        path = Path(vertices, closed=True)
                        xin, yin = np.meshgrid(raster.x, raster.y)
                        xy_in = np.vstack([xin.flatten(), yin.flatten()]).T
                        ridxs = np.where(
                            np.logical_and(
                                np.logical_and(
                                    xy_in[:, 0] >= np.min(path.vertices[:, 0]),
                                    xy_in[:, 0] <= np.max(path.vertices[:, 0])
                                    ),
                                np.logical_and(
                                    xy_in[:, 1] >= np.min(path.vertices[:, 1]),
                                    xy_in[:, 1] <= np.max(path.vertices[:, 1]))
                                ))
                        ridxs_mask = path.contains_points(xy_in[ridxs])
                        rvalues = raster.values.flatten()
                        rvalues = rvalues[ridxs][np.where(ridxs_mask)]
                        if self._initial_values[idx] < 0:
                            self.mesh._values[i] = np.min(rvalues)
                        elif self._initial_values[idx] > 0:
                            self.mesh._values[i] = np.max(rvalues)
                        else:
                            raise Exception('unreachable')

            else:
                msg = 'duck-type error for anti aliasing method '
                msg += f'{self._args.anti_aliasing_method}'
                raise Exception(msg)

    def _expand_tile_index(self, path):
        validate_tile_index(path)
        raster_paths = []
        with fiona.open(path) as src:
            for feature in src:
                url = feature['properties']['URL']
                # Check if raster is in database
                res = self._session.query(db.TileIndexRasters).get(url)
                if res is None:
                    tmpfile = request_raster_from_url(url, self._args.verbose)
                    self._put_raster_in_cache(url, tmpfile)
                    res = self._session.query(db.TileIndexRasters).get(url)
                raster_paths.append(self._cache / 'data' / res.name)
        return raster_paths

    def _put_raster_in_cache(self, url, tmpfile):
        print('debug:_put_raster_in_cache()')
        datadir = self._cache / 'data'
        datadir.mkdir(exist_ok=True)
        target_path = datadir / url.split('/')[-1]
        if not target_path.is_file():
            shutil.copyfile(tmpfile.name, target_path)
    #         self._validate_raster_local(
    #             Raster(target_path), tmpraster.md5
    #     os.copyfile()
    #     tgtraster.save(target_path)
        raster = Raster(target_path)
        bbox = raster.bbox
        # pylint: disable=no-member
        geom = box(
                bbox.xmin,
                bbox.ymin,
                bbox.xmax,
                bbox.ymax
                )
        geom = transform_polygon(geom, raster.crs, 4326)
    #     md5 = raster.md5
        self._session.add(db.TileIndexRasters(
            geom=from_shape(
                geom,
                srid=4326
                ),
            url=url,
            name=target_path.name,
            md5=raster.md5))
        self._session.commit()
    #     return target_path, raster.m
    #     breakpoint()

    @property
    @lru_cache(maxsize=None)
    def _rasters(self):
        raster_paths = []
        for path in self._args.DEM:
            path = pathlib.Path(path)
            if not path.is_file():
                msg = f"No such file: {str(path)}"
                raise FileNotFoundError(msg)
            if check_if_uri_is_tile_index(path):
                raster_paths.extend(self._expand_tile_index(path))
            else:
                raster_paths.append(path)
        return raster_paths

    @property
    @lru_cache(maxsize=None)
    def _pyenv(self):
        return pathlib.Path("/".join(sys.executable.split('/')[:-2]))

    @property
    @lru_cache(maxsize=None)
    def _cache(self):
        cache = self._pyenv / '.cache'
        cache.mkdir(exist_ok=True)
        return cache

    @property
    @lru_cache(maxsize=None)
    def _session(self):
        return db.spatialite_session(self._cache / 'index.db', echo=False)


def polygon_sort(corners):
    # calculate centroid of the polygon
    n = len(corners)  # of corners
    cx = float(sum(x for x, y in corners)) / n
    cy = float(sum(y for x, y in corners)) / n
    # create a new list of corners which includes angles
    cornersWithAngles = []
    for x, y in corners:
        an = (np.arctan2(y - cy, x - cx) + 2.0 * np.pi) % (2.0 * np.pi)
        cornersWithAngles.append((x, y, an))
    # sort it using the angles
    cornersWithAngles.sort(key=lambda tup: tup[2])
    # return the sorted corners w/ angles removed
    return [(x, y) for x, y, _ in cornersWithAngles]


def check_if_uri_is_tile_index(uri):
    try:
        fiona.open(uri, 'r')
        return True
    except fiona.errors.DriverError:
        return False


def validate_tile_index(path):
    with fiona.open(path) as src:
        for feature in src:
            url = feature['properties'].get("URL")
            if url is None:
                raise AttributeError(
                    f"No 'URL' entry for feature with id {feature['id']} "
                    f"on file {path}")


def request_raster_from_url(url, verbose=False):
    """ returns :class:`tempfile.NamedTemporaryFile` object """
    # pylint: disable=R1732
    tmpfile = tempfile.NamedTemporaryFile()
    with open(tmpfile.name, 'wb') as f:
        response = requests.get(url, stream=True)
        total = response.headers.get('content-length')

        if total is None:
            f.write(response.content)
        else:
            downloaded = 0
            total = int(total)
            for data in response.iter_content(
                    chunk_size=max(int(total/1000), 1024*1024)):
                downloaded += len(data)
                f.write(data)
                done = int(50*downloaded/total)
                if verbose:
                    sys.stdout.write(
                        '\r[{}{}]'.format('█' * done, '.' * (50-done)))
                    sys.stdout.flush()
    if verbose:
        sys.stdout.write('\n')
    return tmpfile


def transform_polygon(polygon, src_crs, dst_crs):
    src_crs = CRS.from_user_input(src_crs)
    dst_crs = CRS.from_user_input(dst_crs)
    if not dst_crs.equals(src_crs):
        transformer = Transformer.from_crs(
            src_crs, dst_crs, always_xy=True)
        polygon = transform(transformer.transform, polygon)
    return polygon


def validate_raster_local(raster, md5):
    if raster.md5 != md5:
        raise Exception(f'Checksum mismatch for path {str(raster.path)}')


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('input_mesh_path', help="Path to input mesh.")
    parser.add_argument('output_mesh_path', help="Path to output mesh.")
    parser.add_argument(
        'DEM', nargs="+",
        help="List of at least one input DEM(s). "
        "These will be interpolated in the same order as "
        "they are given. Recommended is to pass them in order of lowest "
        "priority to highest priority."
        )
    parser.add_argument(
        "--overwrite", action="store_true",
        help="Used in case the output_mesh_path exists and the user wants to "
        "allow overwrite.")
    parser.add_argument(
        "--nprocs", type=int, help="Total number of processors to use. This "
        "algorithm can make use of virtual cores, so this value is not "
        "restricted to the number of physical cores or the amount of DEM's "
        "to be interpolated.")
    parser.add_argument(
        "--chunk-size", type=int, help="Useful when passing large rasters that"
        " do not fit in memory, it will subdivide the rasters into boxes "
        " of maximum pixel size chunk-size x chunk-size. If your job runs out "
        "of memory, try using --chunk-size=3000")
    parser.add_argument(
        "--crs",
        help="Input mesh CRS. Output will have the same CRS as the input.")
    parser.add_argument("--verbose", action="store_true")
    parser.add_argument("--use-anti-aliasing", action="store_true")
    parser.add_argument(
        "--anti-aliasing-method",
        default="reuse",
        choices=['reuse', 'fv']
        )
    return parser.parse_args()


def main():
    Interp(parse_args()).main()


if __name__ == '__main__':
    main()



import logging
import json
import pathlib
import os
from functools import lru_cache
import hashlib
from multiprocessing import Pool

from pyproj import CRS
from shapely import ops
from shapely.geometry import MultiPolygon, box
import fiona
import geoalchemy2

from ocsmesh import Hfun, Geom, Raster, db


class _ConfigManager:

    def __init__(self, args, session):
        self._args = args
        self._session = session
        self._certify_config()

    def get_geom(self):
        self._logger.debug('get_geom()')

        geom_collection = []

        if self._geom_nprocs is not None:
            job_args = []
            hashes = []
            geom = None

        for id, geom_opts in self._geom.items():
            self._logger.debug(f'get_geom(): processsing group id={id}')
            zmin = geom_opts.get("zmin")
            zmax = geom_opts.get("zmax")
            driver = geom_opts.get("driver", "matplotlib")

            for raster_path, raster_opts in self._get_raster_by_id(id):
                self._logger.debug(
                    f'get_geom(): appending raster {raster_path} for '
                    'parallel processing.')
                hash = _geom_identifier(
                    zmin, zmax, driver, Raster(raster_path).md5)
                query = self._session.query(db.GeomCollection).get(hash)
                if query is None:
                    chunk_size = raster_opts.get("chunk_size")
                    if self._geom_nprocs is not None:
                        if chunk_size == 0:
                            job_args.append((
                                raster_path, raster_opts, zmin, zmax,
                                geom_opts.get("join_method"),
                                driver,
                                chunk_size,
                                raster_opts.get("overlap", 2)))
                            hashes.append(hash)
                        else:
                            geom = _geom_raster_processing_worker(
                                raster_path, raster_opts, zmin, zmax,
                                geom_opts.get("join_method"),
                                driver, chunk_size,
                                raster_opts.get("overlap"))
                    else:
                        geom = _geom_raster_processing_worker(
                            raster_path, raster_opts, zmin, zmax,
                            geom_opts.get("join_method"),
                            driver, chunk_size,
                            raster_opts.get("overlap"))
                        self._save_geom_to_db(
                            geom, raster_path.name, zmin, zmax, driver, hash)
                        self._session.commit()
                else:
                    geom = Geom(
                        geoalchemy2.shape.to_shape(query.geom),
                        crs=self._crs)

                if geom is not None:
                    geom_collection.append(geom)

        if self._geom_nprocs is not None:
            self._logger.debug(
                'get_geom(): executing parallel geom computations...')
            with Pool(processes=self._geom_nprocs) as pool:
                res = pool.starmap(_geom_raster_processing_worker, job_args)
            pool.join()
            for i, geom in enumerate(res):
                geom_collection.append(geom)
                self._save_geom_to_db(
                    geom, job_args[i][0].name, job_args[i][2],
                    job_args[i][3], job_args[i][5], hashes[i])
            self._session.commit()
            del res

        for feature in self._features:
            raise NotImplementedError('features')

        mpc = []
        for geom in geom_collection:
            mpc.append(geom.multipolygon)
        self._logger.debug('get_geom(): apply unary_union...')
        mp = ops.unary_union(mpc)

        return Geom(mp, crs=self._crs)

    def get_hfun(self, geom=None):
        self._logger.debug('get_hfun()')

        if 'rasters' in self._hfun:
            for id, hfun_raster_opts in self._hfun['rasters'].items():
                self._logger.debug(f'get_hfun(): processsing group id={id}')

                for raster_path, raster_opts in self._get_raster_by_id(id):
                    self._logger.debug(
                        f'get_hfun(): appending raster {raster_path} for '
                        'parallel processing.')
                    raster = Raster(raster_path, crs=raster_opts.get('crs'))
                    _apply_raster_opts(raster, raster_opts)
                    hfun = Hfun(
                        raster,
                        hmin=self._hfun_hmin,
                        hmax=self._hfun_hmax,
                        nprocs=self._hfun_nprocs)
                    _apply_hfun_raster_opts(hfun, hfun_raster_opts)
                    mesh = hfun.get_mesh(geom=geom)

        if 'features' in self._hfun:
            raise NotImplementedError("config.hfun.features not implemented")

        return Hfun(mesh, crs=self._crs)

    def _certify_config(self):
        # pylint: disable=W0104

        self._config
        self._rasters
        self._features
        self._geom
        self._hfun
        self._outputs
        self._logger.debug(" done checking configuration file")

    def _get_raster_by_id(self, rast_id):

        def check_if_uri_is_tile_index(uri):
            try:
                fiona.open(uri, 'r')
                return True
            except fiona.errors.DriverError:
                return False

        raster_opts = self._rasters[rast_id]

        if 'http' in raster_opts['uri'] or 'ftp' in raster_opts['uri']:
            raise NotImplementedError("URI is internet address")

        uri = pathlib.Path(os.path.expandvars(raster_opts['uri']))
        uri = pathlib.Path(self._path).parent / uri

        if not uri.is_file():
            raise FileNotFoundError(f"No file with path: {str(uri.resolve())}")

        if check_if_uri_is_tile_index(uri):
            raise NotImplementedError('URI is a tile index')

        chunk_size = raster_opts.get("chunk_size")
        if chunk_size is None:
            chunk_size = self._config.get("chunk_size")
        if chunk_size is None:
            chunk_size = self._args.chunk_size
        raster_opts.update({"chunk_size": chunk_size})
        return [(uri, raster_opts)]

    def _save_geom_to_db(self, geom, source, zmin, zmax, driver, key):
        self._logger.debug("_save_geom_to_db()")
        _original_crs = geom.crs
        if not _original_crs.equals(CRS.from_epsg(4326)):
            self._logger.debug(f"tranforming from {geom.crs} to EPSG:4326")
            geom.transform_to('EPSG:4326')
        self._session.add(
            db.GeomCollection(
                geom=geoalchemy2.shape.from_shape(geom.multipolygon),
                source=source,
                zmin=zmin,
                zmax=zmax,
                driver=driver,
                id=key))
        if not geom.crs.equals(_original_crs):
            geom.transform_to(_original_crs)

    @property
    @lru_cache(maxsize=None)
    def _config(self):
        with open(self._path, 'r') as f:
            config = json.loads(f.read())

        if not isinstance(config, dict):
            raise TypeError('config json must be a dictionary.')

        if not any(x in config.keys() for x in ['rasters', 'features']):
            raise KeyError(
                "Configuration file must contain at least one of 'rasters' "
                "or 'features' keys.")

        return config

    @property
    def _path(self):
        return self._args.config_file

    @property
    @lru_cache(maxsize=None)
    def _crs(self):
        return CRS.from_user_input(self._config.get("crs", "EPSG:4326"))

    @property
    @lru_cache(maxsize=None)
    def _rasters(self):

        config_rasters = self._config.get('rasters')

        config_rasters = {} if config_rasters is None else config_rasters

        if not isinstance(config_rasters, (dict, list)):
            raise TypeError(
                "config.rasters must be a dictionary or list of "
                "dictionaries containing at least one 'id' and one 'uri' "
                "key.")

        if isinstance(config_rasters, dict):
            config_rasters = [config_rasters]

        _config_rasters = {}

        for i, config_raster in enumerate(config_rasters):

            if not isinstance(config_raster, dict):
                raise TypeError(
                    "config.rasters must be a dictionary or list of "
                    "dictionaries containing at least one 'id' and one "
                    "'uri' key.")

            config_raster_id = config_raster.get('id')

            if config_raster_id is None:
                raise KeyError(
                    "config.rasters entry must contain a unique 'id' key.")

            if config_raster_id in _config_rasters:
                raise KeyError(
                    "'id' entry in config.raster must be unique. "
                    f"repeated key: {config_raster_id}")

            config_raster_uri = config_raster.get('uri')

            if config_raster_uri is None:
                raise KeyError(
                    "config.rasters entry must contain a 'uri' key "
                    "(mutually exclusive).")

            if 'http' in config_raster_uri or 'ftp' in config_raster_uri:
                raise NotImplementedError(
                    "URI provided is an internet address.")

            opts = config_rasters[i].copy()
            opts.pop('id')
            _config_rasters.update({config_raster_id: opts})

        return _config_rasters

    @property
    @lru_cache(maxsize=None)
    def _features(self):
        config_features = self._config.get('features')
        # _features = [] if config_features is None else config_features
        if config_features is not None:
            raise NotImplementedError("config.features is not yet implemented")
        return []

    @property
    @lru_cache(maxsize=None)
    def _geom(self):
        config_geom = self._config.get("geom")
        if config_geom is None:
            raise KeyError("Configuration file must contain a 'geom' key.")
        _geom = {}
        if not isinstance(config_geom, dict):
            raise TypeError(
                "config.geom must be a dictionary containing 'rasters' or "
                "'features' keys.")

        if not any(
                x in config_geom.keys() for x in ['rasters', 'features']):
            raise TypeError(
                "config.geom must be a dictionary containing 'rasters' or "
                "'features' keys.")

        if 'rasters' in config_geom:
            config_geom_rasters = config_geom["rasters"].copy()
            if not isinstance(config_geom_rasters, (dict, list)):
                raise TypeError(
                    "geom.rasters must be a dictionary or list of dictionaries"
                    " containing at least one 'id' key that matches an some "
                    "id on the 'rasters' key.")
            if isinstance(config_geom_rasters, dict):
                config_geom_rasters = [config_geom_rasters]
            for geom_raster in config_geom_rasters:
                geom_raster_id = geom_raster.pop('id')
                if geom_raster_id in self._rasters:
                    _geom.update({geom_raster_id: geom_raster})
                else:
                    raise KeyError(
                        f'No raster with id={geom_raster_id} specified in '
                        'config.rasters')

        if 'features' in config_geom:
            raise NotImplementedError(
                'config.geom.features not yet implemented')

        return _geom

    @property
    @lru_cache(maxsize=None)
    def _hfun(self):
        config_hfun = self._config.get("hfun")
        if config_hfun is None:
            raise KeyError("Configuration file must contain a 'hfun' key.")

        if not isinstance(config_hfun, (int, float, dict)):
            raise TypeError(
                "config.geom must be an scalar value (constant size) or a "
                "dictionary containing either the 'rasters' or 'features' "
                "keys, or both.")

        if isinstance(config_hfun, (int, float)):
            raise NotImplementedError('Constant size funtion.')

        _config_hfun = {}
        _config_hfun.update({"hmin": config_hfun.get("hmin")})
        _config_hfun.update({"hmax": config_hfun.get("hmax")})

        _config_hfun_rasters = config_hfun.get("rasters").copy()
        if _config_hfun_rasters is not None:
            _config_hfun.update({"rasters": {}})
            if not isinstance(_config_hfun_rasters, (list, dict)):
                raise TypeError(
                    'config.hfun.rasters must be a dictionary or list of '
                    'dictionaries that contain an "id" key that matches a '
                    'key from the config.rasters entry.')
            if isinstance(_config_hfun_rasters, dict):
                _config_hfun_rasters = [_config_hfun_rasters]

            for hfun_raster_opts in _config_hfun_rasters:
                hfun_raster_id = hfun_raster_opts.pop('id')
                if hfun_raster_id not in self._rasters.keys():
                    raise KeyError(
                        f'No raster with id={hfun_raster_id} specified in '
                        'config.rasters')
                contours = hfun_raster_opts.get("contours", [])
                if isinstance(contours, dict):
                    contours = [contours]
                hfun_raster_opts.update({"contours": contours})
                _config_hfun["rasters"].update(
                        {hfun_raster_id: hfun_raster_opts})

                # _geom.update({geom_raster_id: geom_raster})

        if 'features' in config_hfun:
            raise NotImplementedError(
                'config.hfun.features not yet implemented')

        return _config_hfun

    @property
    def _outputs(self):
        config_outputs = self._config.get("outputs")
        _outputs = {}
        return _outputs

    @property
    @lru_cache(maxsize=None)
    def _geom_nprocs(self):
        nprocs = self._config["geom"].get("nprocs")
        if nprocs is None:
            nprocs = self._config.get("nprocs")
        if nprocs is None:
            nprocs = self._args.geom_nprocs
        if nprocs is None:
            nprocs = self._args.nprocs
        return nprocs

    @property
    @lru_cache(maxsize=None)
    def _hfun_nprocs(self):
        nprocs = self._config["hfun"].get("nprocs")
        if nprocs is None:
            nprocs = self._config.get("nprocs")
        if nprocs is None:
            nprocs = self._args.hfun_nprocs
        if nprocs is None:
            nprocs = self._args.nprocs
        return nprocs

    @property
    def _hfun_hmin(self):
        return self._config["hfun"].get("hmin")

    @property
    def _hfun_hmax(self):
        return self._config["hfun"].get("hmax")

    @property
    def _hfun_raster_opts(self):
        opts = self._config["hfun"]["rasters"]

    @property
    @lru_cache(maxsize=None)
    def _logger(self):
        return logging.getLogger(__name__ + '.' + self.__class__.__name__)


def _geom_identifier(zmin, zmax, driver, salt):
    zmin = "" if zmin is None else f"{zmin:G}"
    zmax = "" if zmax is None else f"{zmax:G}"
    return hashlib.md5(
            f"{zmin}{zmax}{driver}{salt}".encode('utf-8')
            ).hexdigest()


def _hfun_identifier(config_hfun):
    pass


def _apply_raster_opts(raster, raster_opts):
    for key, opt in raster_opts.items():
        if key == "resample":
            if isinstance(opt, (float, int)):
                raster.resample(opt)
            else:
                raster.resample(**opt)
        if key == 'fill_nodata':
            raster.fill_nodata()
        if key == 'clip':
            if isinstance(opt, dict):
                raster.clip(
                    MultiPolygon(
                        [box(
                            opt['xmin'],
                            opt['ymin'],
                            opt['xmax'],
                            opt['ymax'])]))
            else:
                raise NotImplementedError("clip by geometry")
        if key == 'chunk_size':
            raster.chunk_size = opt

        if key == 'overlap':
            raster.overlap = opt


def _apply_hfun_raster_opts(hfun, hfun_raster_opts):
    for key, opts in hfun_raster_opts.items():
        if key == 'contours':
            for kwargs in opts:
                hfun.add_contour(**kwargs)
        if key == 'features':
            for kwargs in opts:
                hfun.add_features(**kwargs)
        if key == 'subtidal_flow_limiter':
            if not isinstance(opts, (bool, dict)):
                raise TypeError(
                    "subtidal_flow_limiter options must be "
                    "a boolean or dict.")
            if opts is True:
                hfun.add_subtidal_flow_limiter()
            else:
                hfun.add_subtidal_flow_limiter(**opts)


def _geom_raster_processing_worker(
        raster_path,
        raster_opts,
        zmin,
        zmax,
        join_method,
        driver,
        chunk_size,
        overlap,
):
    raster = Raster(raster_path)
    _apply_raster_opts(raster, raster_opts)
    geom = Geom(raster.get_multipolygon(
        zmin=zmin,
        zmax=zmax),
        join_method=join_method,
        driver=driver,
        nprocs=1)

    return geom



#! python
import unittest
from copy import deepcopy
import numpy as np
import ocsmesh


class SizeFromMesh(unittest.TestCase):

    def setUp(self):
        rast = ocsmesh.raster.Raster('test_dem.tif')

        hfun_orig = ocsmesh.hfun.hfun.Hfun(rast, hmin=100, hmax=1500)
        hfun_orig.add_contour(level=0, expansion_rate=0.001, target_size=100)
        hfun_orig_jig = hfun_orig.msh_t()

        self.hfun_orig_val = hfun_orig_jig.value

        hfun_calc_jig = deepcopy(hfun_orig_jig)
        mesh_calc = ocsmesh.mesh.mesh.Mesh(hfun_calc_jig)
        self.hfun_calc = ocsmesh.hfun.hfun.Hfun(mesh_calc)
        self.hfun_calc.size_from_mesh()

    def test_calculated_size(self):
        hfun_calc_jig = self.hfun_calc.msh_t()

        hfun_calc_val = hfun_calc_jig.value
        hfun_val_diff = self.hfun_orig_val - hfun_calc_val

        # TODO: Come up with a more robust criteria
        threshold = 0.2
        err_value = np.max(np.abs(hfun_val_diff))/np.max(self.hfun_orig_val)
        self.assertTrue(err_value < threshold)

if __name__ == '__main__':
    unittest.main()



"""This module defines a base type for mesh objects.
"""

import numpy as np
import numpy.typing as npt
from jigsawpy import jigsaw_msh_t

class BaseMesh:
    """Base class for mesh types in OCSMesh.

    Attributes
    ----------
    msh_t : jigsaw_msh_t
        Refernece to underlying jigsaw mesh object.
    coord : array-like
        Coordinates of mesh nodes.

    Methods
    -------
    """

    @property
    def msh_t(self) -> jigsaw_msh_t:
        """Read-only property returning reference to the jigsawpy mesh object

        Notes
        -----
        The property is read-only, however the returned value is not.
        That means if the user mutates the returned jigsaw_msh_t
        object, it will affect the mesh.
        """

        return self._msh_t

    @property
    def coord(self) -> npt.NDArray[np.float32]:
        """Read-only property for coordinates of mesh points

        Raises
        ------
        If the number of mesh dimensions is not 2 or 3.

        Notes
        -----
        The property is read-only, however the returned value is not.
        That means if the user mutates the returned numpy array
        object, it will affect the mesh.
        """

        if self.msh_t.ndims == 2: # pylint: disable=R1705
            return self.msh_t.vert2['coord']
        elif self.msh_t.ndims == 3:
            return self.msh_t.vert3['coord']

        raise ValueError(f'Unhandled mesh dimensions {self.msh_t.ndims}.')



#! /usr/bin/env python
import setuptools
import subprocess
import distutils.cmd
import shutil
import platform
from multiprocessing import cpu_count
from pathlib import Path
import sys
import os

PARENT = Path(__file__).parent.absolute()
PYENV_PREFIX = Path("/".join(sys.executable.split('/')[:-2]))
SYSLIB = {
    "Windows": "jigsaw.dll",
    "Linux": "libjigsaw.so",
    "Darwin": "libjigsaw.dylib"}

if "install_jigsaw" not in sys.argv:
    if "develop" not in sys.argv:
        if "install" in sys.argv:
            libsaw = PYENV_PREFIX / 'lib' / SYSLIB[platform.system()]
            if not libsaw.is_file():
                subprocess.check_call(
                    [sys.executable, "setup.py", "install_jigsaw"])


class InstallJigsawCommand(distutils.cmd.Command):
    """Custom build command."""

    user_options = []

    def initialize_options(self): pass

    def finalize_options(self): pass

    def run(self):
        self.announce('Loading JIGSAWPY from GitHub', level=3)
        # init jigsaw-python submodule
        subprocess.check_call(
            ["git", "submodule", "update",
             "--init", "submodules/jigsaw-python"])
        # install jigsawpy
        os.chdir(PARENT / 'submodules/jigsaw-python')
        subprocess.check_call(["git", "checkout", "master"])
        self.announce('INSTALLING JIGSAWPY', level=3)
        subprocess.check_call(["python", "setup.py", "install"])
        # install jigsaw
        self.announce(
            'INSTALLING JIGSAW LIBRARY AND BINARIES FROM '
            'https://github.com/dengwirda/jigsaw-python', level=3)
        os.chdir("external/jigsaw")
        os.makedirs("build", exist_ok=True)
        os.chdir("build")
        gcc, cpp = self._check_gcc_version()
        subprocess.check_call(
            ["cmake", "..",
             "-DCMAKE_BUILD_TYPE=Release",
             f"-DCMAKE_INSTALL_PREFIX={PYENV_PREFIX}",
             f"-DCMAKE_C_COMPILER={gcc}",
             f"-DCMAKE_CXX_COMPILER={cpp}",
             ])
        subprocess.check_call(["make", f"-j{cpu_count()}", "install"])
        libsaw_prefix = list(PYENV_PREFIX.glob("**/*jigsawpy*")).pop() / '_lib'
        os.makedirs(libsaw_prefix, exist_ok=True)
        envlib = PYENV_PREFIX / 'lib' / SYSLIB[platform.system()]
        os.symlink(envlib, libsaw_prefix / envlib.name)
        os.chdir(PARENT)
        subprocess.check_call(
          ["git", "submodule", "deinit", "-f", "submodules/jigsaw-python"])

    def _check_gcc_version(self):
        cpp = shutil.which("c++")
        major, minor, patch = subprocess.check_output(
            [cpp, "--version"]
            ).decode('utf-8').split('\n')[0].split()[-1].split('.')
        current_version = float(f"{major}.{minor}")
        if current_version < 7.:
            raise Exception(
                'JIGSAW requires GCC version 7 or later, got '
                f'{major}.{minor}.{patch} from {cpp}')
        return shutil.which("gcc"), cpp


setuptools.setup(
    cmdclass={
        'install_jigsaw': InstallJigsawCommand,
        },
)



"""This module defines the size function factory class

Instead of importing specific size function types, users can import
factory `Hfun` object from this module which creates the correct
type of size function based on the input arguments passed to it
"""

from typing import Any

from ocsmesh.hfun.base import BaseHfun
from ocsmesh.hfun.raster import HfunRaster
from ocsmesh.hfun.mesh import HfunMesh
from ocsmesh.mesh.mesh import EuclideanMesh2D
from ocsmesh.raster import Raster
from ocsmesh.hfun.collector import HfunCollector, CanCreateHfun


class Hfun:
    """ Size function object factory

    Factory class that creates and returns concrete mesh size function
    object based on the type of the inputs.

    Methods
    -------
    is_valid_type(hfun_object)
        Static method to check if an object is a valid size function
        type

    Notes
    -----
    The object created when using this class is not an instance of
    this class or a subclass of it. It is a subclass of BaseHfun
    instead.
    """


    # pylint: disable=no-member

    def __new__(cls, hfun: CanCreateHfun, **kwargs: Any) -> BaseHfun:
        """
        Parameters
        ----------
        hfun : CanCreateHfun
            Object to create the mesh size function from. The type of
            this object determines the created size function object
            type
        **kwargs : dict, optional
            Keyword arguments passed to the constructor of the
            correct size function type
        """

        if isinstance(hfun, Raster): # pylint: disable=R1705
            return HfunRaster(hfun, **kwargs)

        elif isinstance(hfun, EuclideanMesh2D):
            return HfunMesh(hfun)

        elif isinstance(hfun, (list, tuple)):
            return HfunCollector(hfun, **kwargs)

        else:
            raise TypeError(
                f'Argument hfun has the wrong type, type {type(hfun)}.')

    @staticmethod
    def is_valid_type(hfun_object: Any) -> bool:
        """Checks if an object is a valid size function type"""
        return isinstance(hfun_object, BaseHfun)



#!/usr/bin/env python
import sys
import gc
import logging
from pathlib import Path
from copy import deepcopy

import jigsawpy
import numpy as np
import geopandas as gpd
from shapely.geometry import Polygon, MultiPolygon

from ocsmesh import Raster, Geom, Hfun, Mesh
from ocsmesh import utils


logging.basicConfig(
    stream=sys.stdout,
    format='%(asctime)s,%(msecs)d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
    datefmt='%Y-%m-%d:%H:%M:%S'
    )
#logging.getLogger().setLevel(logging.DEBUG)
#logging.getLogger().setLevel(logging.INFO)

_logger = logging.getLogger(__name__)

class RemeshByDEM:

    @property
    def script_name(self):
        return 'remesh_by_dem'

    def __init__(self, sub_parser):

        # e.g
        # ./remesh \
        #   --mesh <MESH_ADD> \
        #   --output <OUT_FILE_ADDR> \
        #   --hmin <GLOBAL_MIN_SIZE> \
        #   --contour <LEVEL0 [EXPAND0 SIZE0]> \
        #   --contour <LEVEL1 [EXPAND1 SIZE1]> \
        #   --contour <LEVELN [EXPANDN SIZEN]> \
        #   --constant <ABOVELEVELA SIZEA> \
        #   --constant <ABOVELEVELB SIZEB> \
        #   --zmax <MAX_ELEV_DEM> \
        #   <DEM_FILES>*.tif

        this_parser = sub_parser.add_parser(
            self.script_name,
            help="Locally refine an existing input mesh based on input DEMs")

        this_parser.add_argument('--mesh', required=True, type=Path)
        this_parser.add_argument('--mesh-crs', default='EPSG:4326')


        this_parser.add_argument(
            '--contour',
            action='append', nargs='+', type=float, dest='contours',
            metavar='CONTOUR_DEFN', default=[],
            help="Each contour's (level, [expansion, target])"
                 " to be applied on all size functions in collector")
        this_parser.add_argument(
            '--constant',
            action='append', nargs=2, type=float, dest='constants',
            metavar='CONST_DEFN', default=[],
            help="Specify constant mesh size above a given contour level"
                 " by passing (lower_bound, target_size) for each constant")
        this_parser.add_argument('--hmin', type=float, default=250)
        this_parser.add_argument('--zmax', type=float, default=0)
        this_parser.add_argument(
            '--clip-by-base', action='store_true',
            help='Flag to clip input DEMs using base mesh polygon')

        this_parser.add_argument('--geom', type=Path)
        this_parser.add_argument('--hfun', type=Path)
        this_parser.add_argument('--hfun-crs', default='EPSG:4326')

        this_parser.add_argument('-s', '--sieve', type=float)
        this_parser.add_argument(
            '--interpolate', nargs='+', type=Path, default=[],
            help="To interpolate from depth of DEMs not involved in"
                 " the remeshing process")

        this_parser.add_argument('-o', '--output', type=Path)
        this_parser.add_argument('-f', '--output-format', default='2dm')
        this_parser.add_argument('-k', '--keep-intermediate', action='store_true')
        this_parser.add_argument('--nprocs', type=int, default=-1)

        this_parser.add_argument('dem', nargs='+', type=Path)


    @staticmethod
    def _read_geom_hfun(geom_file, hfun_file, hfun_crs):
        _logger.info("Read geom and hfun from disk")
        _logger.info("Readng geometry...")
        gdf_geom = gpd.read_file(geom_file)
        poly_list = []
        for i in gdf_geom.geometry:
            if isinstance(i, MultiPolygon):
                poly_list.extend(i)
            elif isinstance(i, Polygon):
                poly_list.append(i)
        geom = Geom(MultiPolygon(poly_list), crs=gdf_geom.crs)
        _logger.info("Done")

        _logger.info("Readng size function...")
        if hfun_crs is None:
            hfun_crs = "EPSG:4326"
        hfun = Hfun(Mesh.open(hfun_file, crs=hfun_crs))
        _logger.info("Done")

        return geom, hfun


    def run(self, args):

        # Get inputs
        base_path = args.mesh
        mesh_crs = args.mesh_crs

        dem_paths = args.dem
        contours = args.contours
        constants = args.constants
        hmin = args.hmin
        zmax = args.zmax
        clip_by_mesh = args.clip_by_base

        geom_file = args.geom
        hfun_file = args.hfun
        hfun_crs = args.hfun_crs

        sieve = args.sieve
        interp = args.interpolate

        out_path = args.output
        out_format = args.output_format
        write_intermediate = args.keep_intermediate
        nprocs = args.nprocs

        # Process inputs
        contour_defns = []
        for contour in contours:
            if len(contour) > 3:
                raise ValueError(
                    "Invalid format for contour specification."
                    " It should be level [expansion target-size].")
            level, expansion_rate, target_size = [
                    *contour, *[None]*(3-len(contour))]
            contour_defns.append((level, expansion_rate, target_size))

        constant_defns = []
        for lower_bound, target_size in constants:
            constant_defns.append((lower_bound, target_size))


        if out_path is None:
            out_path = base_path.parent / ('remeshed.' + out_format)
        out_path.parent.mkdir(exist_ok=True, parents=True)

        nprocs = -1 if nprocs is None else nprocs

        geom_rast_list = []
        hfun_rast_list = []
        interp_rast_list = []

        # Low priority interpolation. e.g. user remeshes based on NCEI
        # but still wants to interpolate GEBCO on mesh
        for dem in interp:
            interp_rast_list.append(Raster(dem))

        # NOTE: Region of interest is calculated from geom_rast_list
        # so they're needed even if hfun and geom are read from file
        _logger.info("Read DEM files")
        for dem_path in dem_paths:
            geom_rast_list.append(Raster(dem_path))
            hfun_rast_list.append(Raster(dem_path))
            interp_rast_list.append(Raster(dem_path))

        _logger.info("Read base mesh")
        if mesh_crs is None:
            mesh_crs = "EPSG:4326"
        init_mesh = Mesh.open(str(base_path), crs=mesh_crs)
        # TODO: Cleanup isolates?

        log_calculation = True
        # Read geometry and hfun from files if provided
        if (geom_file and hfun_file
                and geom_file.is_file() and hfun_file.is_file()):
            geom, hfun = self._read_geom_hfun(geom_file, hfun_file, hfun_crs)
            log_calculation = False

        # Create geometry and hfun from inputs
        elif dem_paths:

            _logger.info("Creating geometry object")
            geom_base_mesh = None
            geom_inputs = geom_rast_list
            if clip_by_mesh:
                _logger.info("Clip raster data by base mesh")
                geom_base_mesh = deepcopy(init_mesh)
            else:
                _logger.info("Union raster data with base mesh")
                geom_inputs = [deepcopy(init_mesh), *geom_rast_list]
            geom = Geom(
                    geom_inputs,
                    base_mesh=geom_base_mesh,
                    zmax=zmax,
                    nprocs=nprocs)


            # NOTE: Instead of passing base mesh to be used as boundary,
            # it is passed as an hfun itself
            hfun_base_mesh = Hfun(deepcopy(init_mesh))
            _logger.info("Calculating size function from mesh...")
            hfun_base_mesh.size_from_mesh()
            _logger.info("Done")
            hfun = Hfun(
                [hfun_base_mesh, *hfun_rast_list],
                hmin=hmin,
                hmax=np.max(hfun_base_mesh.msh_t().value),
                nprocs=nprocs)

            for level, expansion_rate, target_size in contour_defns:
                if expansion_rate is None:
                    expansion_rate = 0.1
                if target_size is None:
                    target_size = hmin
                _logger.info(f"Adding contour to collector:"
                             f" {level} {expansion_rate} {target_size}")
                hfun.add_contour(
                    level, expansion_rate, target_size)

            for lower_bound, target_size in constant_defns:
                hfun.add_constant_value(
                        value=target_size, lower_bound=lower_bound)


            if write_intermediate:
                _logger.info("Calculating final geometry")
                poly_geom = geom.get_multipolygon()

                _logger.info("Writing geom to disk")
                gpd.GeoDataFrame(
                        {'geometry': gpd.GeoSeries(poly_geom)},
                        crs=geom.crs
                    ).to_file(str(out_path)+'.geom.shp')
                del poly_geom
                gc.collect()

                _logger.info("Calculating final size function")
                jig_hfun = hfun.msh_t()

                _logger.info("Writing hfun to disk")
                # This writes in EPSG:4326
                Mesh(jig_hfun).write(
                    str(out_path)+'.hfun.2dm',
                    format='2dm', overwrite=True)
                del jig_hfun
                gc.collect()

                # Read back from file to avoid recalculation of hfun
                # and geom
                geom, hfun = self._read_geom_hfun(
                    str(out_path) + '.geom.shp',
                    str(out_path) + '.hfun.2dm',
                    "EPSG:4326")

                log_calculation = False
        else:
            raise ValueError(
                "Input not valid to initialize geom and hfun")


        if log_calculation:
            # NOTE: If intermediate files are written then we calculated
            # this in previous section
            _logger.info("Calculating final geometry")
        jig_geom = geom.msh_t()
        if log_calculation:
            # NOTE: If intermediate files are written then we calculated
            # this in previous section
            _logger.info("Calculating final size function")
        jig_hfun = hfun.msh_t()

        jig_init = init_mesh.msh_t

        _logger.info("Projecting geometry to be in meters unit")
        utils.msh_t_to_utm(jig_geom)
        _logger.info("Projecting size function to be in meters unit")
        utils.msh_t_to_utm(jig_hfun)
        _logger.info("Projecting initial mesh to be in meters unit")
        utils.msh_t_to_utm(jig_init)


        # pylint: disable=C0325
        if not (jig_geom.crs == jig_hfun.crs == jig_init.crs):
            raise ValueError(
                "Converted UTM CRS for geometry, hfun and init mesh"
                "is not equivalent")


        _logger.info("Calculate remeshing region of interest")
        # Prep for Remeshing
        boxes = [i.get_bbox(crs=jig_geom.crs) for i in geom_rast_list]
        region_of_interest = MultiPolygon(boxes)
        roi_bnds = region_of_interest.bounds
        roi_s = max(roi_bnds[2] - roi_bnds[0], roi_bnds[3] - roi_bnds[1])

        _logger.info("Clip mesh by inverse of region of interest")
        fixed_mesh_w_hole = utils.clip_mesh_by_shape(
            jig_init, region_of_interest, fit_inside=True, inverse=True)

        _logger.info(
                "Get all initial mesh vertices in the region of interest")
        vert_idx_to_refin = utils.get_verts_in_shape(
            jig_hfun, region_of_interest)

        fixed_mesh_w_hole.point['IDtag'][:] = -1
        fixed_mesh_w_hole.edge2['IDtag'][:] = -1

        refine_opts = jigsawpy.jigsaw_jig_t()
        refine_opts.hfun_scal = "absolute"
        refine_opts.hfun_hmin = np.min(jig_hfun.value)
        refine_opts.hfun_hmax = np.max(jig_hfun.value)
        refine_opts.mesh_dims = +2
        # Mesh becomes TOO refined on exact boundaries from DEM
#    refine_opts.mesh_top1 = True
#    refine_opts.geom_feat = True

        jig_remeshed = jigsawpy.jigsaw_msh_t()
        jig_remeshed.ndims = +2

        _logger.info("Remeshing...")
        # Remeshing
        jigsawpy.lib.jigsaw(
                refine_opts,
                jig_geom,
                jig_remeshed,
                init=fixed_mesh_w_hole,
                hfun=jig_hfun)
        jig_remeshed.crs = fixed_mesh_w_hole.crs
        _logger.info("Done")

        if jig_remeshed.tria3['index'].shape[0] == 0:
            _err = 'ERROR: Jigsaw returned empty mesh.'
            _logger.error(_err)
            raise ValueError(_err)

        # TODO: This is irrelevant right now since output file is
        # always is EPSG:4326, enable when APIs for remeshing is added
#    if out_crs is not None:
#        utils.reproject(jig_remeshed, out_crs)

        _logger.info('Finalizing mesh...')
        utils.finalize_mesh(jig_remeshed, sieve)

        _logger.info("Interpolating depths on mesh...")
        # Interpolation
        utils.interpolate_euclidean_mesh_to_euclidean_mesh(
                jig_init, jig_remeshed)
        final_mesh = Mesh(jig_remeshed)
        final_mesh.interpolate(interp_rast_list, nprocs=nprocs)
        _logger.info("Done")


        _logger.info("Writing final mesh to disk...")
        # This writes EPSG:4326 to file, whatever the crs of the object
        final_mesh.write(str(out_path), format=out_format, overwrite=True)
        _logger.info("Done")



import pathlib
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.event import listen
from sqlalchemy.sql import select, func
from sqlalchemy.orm import sessionmaker
from sqlalchemy import create_engine
from sqlalchemy import Column, Float, String
from geoalchemy2 import Geometry
from geoalchemy2 import Raster as _Raster


Base = declarative_base()


class Geom(Base):
    __tablename__ = 'geom'
    id = Column(String, primary_key=True, nullable=False)
    tag = Column(String)
    geom = Column(
        Geometry(
            geometry_type='MULTIPOLYGON',
            management=True
            ),
        nullable=False
        )


class GeomCollection(Base):
    __tablename__ = "geom_collection"
    id = Column(String, primary_key=True, nullable=False)
    source = Column(String)
    zmin = Column(Float)
    zmax = Column(Float)
    driver = Column(String, nullable=False)
    geom = Column(
        Geometry(
            geometry_type='MULTIPOLYGON',
            management=True
            ),
        nullable=False
        )


class Hfun(Base):
    __tablename__ = 'hfun'
    id = Column(String, primary_key=True, nullable=False)
    hfun = Column(
        Geometry(
            geometry_type='MULTIPOLYGON',
            management=True,
            dimension=3,
            ),
        nullable=False
        )


class HfunCollection(Base):
    __tablename__ = 'hfun_collection'
    id = Column(String, primary_key=True, nullable=False)


class Raster(Base):
    __tablename__ = "rasters"
    raster = Column(_Raster, nullable=True)
    md5 = Column(String, primary_key=True)


class TileIndexRasters(Base):
    __tablename__ = 'tile_index_rasters'
    geom = Column(
        Geometry(
            'POLYGON',
            management=True,
            srid=4326
            ),
        nullable=False)
    # raster = Column(Raster(srid=4326, spatial_index=False))
    url = Column(String, primary_key=True, nullable=False)
    name = Column(String, nullable=False)
    md5 = Column(String, nullable=False)


def spatialite_session(path, echo=False):

    def _engine(path, echo=False):
        path = pathlib.Path(path)
        _new_db = not path.is_file()
        engine = create_engine(f'sqlite:///{str(path)}', echo=echo)

        def load_spatialite(dbapi_conn, connection_record):
            dbapi_conn.enable_load_extension(True)
            dbapi_conn.load_extension('mod_spatialite')

        listen(engine, 'connect', load_spatialite)
        if _new_db:
            conn = engine.connect()
            conn.execute(select([func.InitSpatialMetaData()]))
            conn.close()
            Geom.__table__.create(engine)
            GeomCollection.__table__.create(engine)
            Hfun.__table__.create(engine)
            TileIndexRasters.__table__.create(engine)
            HfunCollection.__table__.create(engine)

        return engine

    def _session(engine):
        return sessionmaker(bind=engine)()

    return _session(_engine(path, echo))


def postgis_session():
    raise NotImplementedError



#! /usr/bin/env python

import argparse

from ocsmesh.ops import combine_geometry, combine_hfun
from ocsmesh.cli.cli import CmdCli


class OCSMesh:

    def __init__(self, args, ocsmesh_cli):
        self._args = args
        self._cli = ocsmesh_cli

    def main(self):

        if self._args.command == 'geom':

            nprocs = self._args.nprocs
            if self._args.geom_nprocs:
                nprocs = self._args.geom_nprocs
            nprocs = -1 if nprocs is None else nprocs

            if self._args.geom_cmd == "build":
                arg_dict = dict(
                    dem_files=self._args.dem,
                    out_file=self._args.output,
                    out_format=self._args.output_format,
                    mesh_file=self._args.mesh,
                    ignore_mesh_final_boundary=self._args.ignore_mesh_boundary,
                    zmin=self._args.zmin,
                    zmax=self._args.zmax,
                    chunk_size=self._args.chunk_size,
                    overlap=self._args.overlap,
                    nprocs=nprocs,
                    out_crs=self._args.output_crs,
                    base_crs=self._args.mesh_crs)
                combine_geometry(**arg_dict)

        elif self._args.command == 'hfun':

            nprocs = self._args.nprocs
            if self._args.hfun_nprocs:
                nprocs = self._args.hfun_nprocs
            nprocs = -1 if nprocs is None else nprocs

            if self._args.hfun_cmd == "build":
                arg_dict = dict(
                    dem_files=self._args.dem,
                    out_file=self._args.output,
                    out_format=self._args.output_format,
                    mesh_file=self._args.mesh,
                    hmin=self._args.hmin,
                    hmax=self._args.hmax,
                    contours=self._args.contour,
                    constants=self._args.constants,
                    chunk_size=self._args.chunk_size,
                    overlap=self._args.overlap,
                    method=self._args.method,
                    nprocs=nprocs)
                combine_hfun(**arg_dict)

        elif self._args.command == 'scripts':
            self._cli.execute(self._args)


def create_parser():
    common_parser = argparse.ArgumentParser()

    common_parser.add_argument("--log-level", choices=["info", "debug", "warning"])
    common_parser.add_argument(
        "--nprocs", type=int, help="Number of parallel threads to use when "
        "computing geom and hfun.")
    common_parser.add_argument(
        "--geom-nprocs", type=int, help="Number of processors used when "
        "computing the geom, overrides --nprocs argument.")
    common_parser.add_argument(
        "--hfun-nprocs", type=int, help="Number of processors used when "
        "computing the hfun, overrides --nprocs argument.")
    common_parser.add_argument(
        "--chunk-size",
        help='Size of square window to be used for processing the raster')
    common_parser.add_argument(
        "--overlap",
        help='Size of overlap to be used for between raster windows')

    sub_parse_common = {
            'parents': [common_parser],
            'add_help': False
    }

    parser = argparse.ArgumentParser(**sub_parse_common)
    subp = parser.add_subparsers(dest='command')

    geom_parser = subp.add_parser(
        'geom', **sub_parse_common,
        help="Perform operations related to domain creation and modification.")
    geom_subp = geom_parser.add_subparsers(dest='geom_cmd')
    geom_bld = geom_subp.add_parser(
        'build', **sub_parse_common,
        help="Build command for domain definition")
    geom_bld.add_argument('-o', '--output', required=True)
    geom_bld.add_argument('-f', '--output-format', default="shapefile")
    geom_bld.add_argument('--output-crs', default="EPSG:4326")
    geom_bld.add_argument('--mesh', help='Mesh to extract hull from')
    geom_bld.add_argument(
        '--ignore-mesh-boundary', action='store_true',
        help='Flag to ignore mesh boundary for final boundary union')
    geom_bld.add_argument(
        '--mesh-crs', help='CRS of the input base mesh (overrides)')
    geom_bld.add_argument(
        '--zmin', type=float,
        help='Maximum elevation to consider')
    geom_bld.add_argument(
        '--zmax', type=float,
        help='Maximum elevation to consider')
    geom_bld.add_argument(
        'dem', nargs='+',
        help='Digital elevation model list to be used in geometry creation')

    hfun_parser = subp.add_parser(
        'hfun', **sub_parse_common,
        help="Perform operations related to size function creation and modification.")
    hfun_subp = hfun_parser.add_subparsers(dest='hfun_cmd')
    hfun_bld = hfun_subp.add_parser(
        'build', **sub_parse_common,
        help="Build command for mesh size definition")
    hfun_bld.add_argument('-o', '--output', required=True)
    hfun_bld.add_argument('-f', '--output-format', default="2dm")
    hfun_bld.add_argument('--mesh', help='Base mesh size function')
    hfun_bld.add_argument(
        '--hmax', type=float, help='Maximum element size')
    hfun_bld.add_argument(
        '--hmin', type=float, help='Minimum element size')
    hfun_bld.add_argument(
        '--contour', action='append', nargs='+', type=float, default=[],
        help="Each contour's (level, [expansion, target])"
             " to be applied on all size functions in collector")
    hfun_bld.add_argument(
        '--constant',
        action='append', nargs=2, type=float, dest='constants',
        metavar='CONST_DEFN', default=[],
        help="Specify constant mesh size above a given contour level"
             " by passing (lower_bound, target_size) for each constant")
    hfun_bld.add_argument(
        '--method', type=str, default='exact',
        help='Method used to calculate size function ({exact} |fast)')
    hfun_bld.add_argument(
        'dem', nargs='+',
        help='Digital elevation model list to be used in size function creation')

    # Scripts don't use common arguments as they are standalon code
    scripts_parser = subp.add_parser(
        'scripts',
        help='Access to experimental OCSMesh scripts.')
    cmd_cli = CmdCli(scripts_parser)

    return parser, cmd_cli

def dummy_documentation():
    parser, ocsmesh_cli = create_parser()
    return parser


def main():
    parser, ocsmesh_cli = create_parser()
#    logger.init(args.log_level)
    OCSMesh(parser.parse_args(), ocsmesh_cli).main()


if __name__ == '__main__':
    main()
